"""
Methods for calibration of OMC data.

GNU GPL v3.0
Lukas Adamowicz
V0.1 - April 10, 2019
"""
from numpy import cross, stack, zeros, isnan, sum as nsum, array, mean
from numpy.linalg import norm, solve

from pykinematics.omc import utility
from pykinematics.omc import segmentFrames
from pykinematics import common


def compute_hip_center(pelvis_data, thigh_data, R, origin, marker_names='default', tol=1e-4):
    """
    Compute the hip joint center location using a bias compensated least squares estimation.

    Parameters
    ----------
    pelvis_data : dict
        Dictionary of all the marker data for the pelvis. Each key is a marker name.
    thigh_data : dict
        Dictionary of all the marker data for the thigh. Each key is a marker name
    R : numpy.ndarray
        Rotation matrix or array of rotation matrices representing the rotation from world to local reference frame.
    origin : numpy.ndarray
        Vector of the origin position to use when rotating and transforming the marker positions into a local
        reference frame.
    marker_names : {'default', MarkerNames}, optional
        Either 'default' to use the default marker names, or a MarkerNames object with the names used in pelvis_data
        and thigh_data specified.
    tol : float, optional
        Tolerance for the bias compensation of the joint center. Default is 1e-4.

    Returns
    -------
    c : numpy.ndarray
        Vector from the provided origin to the joint center, expressed in the local reference frame.

    References
    ----------
    Halvorsen, Kjartan.  "Bias compensated least squares estimate of the center of rotation." *J. of Biomech*.
    Vol. 36. 2003.
    Gamage et al. "New least squares solution for estimating the average center of rotation and the axis of rotation."
    *J. of Biomech*. Vol. 35. 2002
    """
    if marker_names == 'default':
        names = utility.MarkerNames()
    else:
        names = marker_names

    # create a mask for NaN values
    mask = ~isnan(R).any(axis=2).any(axis=1)  # check the world to local frame rotation
    mask &= ~isnan(pelvis_data[names.pelvis_c2]).any(axis=1)  # check the thigh cluster origin

    # check all the marker data for nan as all the data has to be the same length
    for mkr in thigh_data.keys():
        mask &= ~isnan(thigh_data[mkr]).any(axis=1)

    # translate and rotate all the thigh marker coordinates into the cluster reference frame
    thigh_rot = dict()
    for mkr in thigh_data.keys():
        thigh_rot[mkr] = (R[mask] @ (thigh_data[mkr][mask]
                                     - origin[mask]).reshape((-1, 3, 1))).reshape((-1, 3))

    n = mask.sum()  # number of samples
    m = len(thigh_rot.keys())  # number of markers

    # create the vectors to store the norms
    norm1 = zeros((m, 3, 1))
    norm2 = zeros((m, 1, 1))
    norm3 = zeros((m, 3, 1))

    for i, mk in enumerate(thigh_rot.keys()):
        norm1[i] = thigh_rot[mk].sum(axis=0).reshape((-1, 3, 1))
        norm2[i] = nsum(thigh_rot[mk].reshape((-1, 1, 3)) @ thigh_rot[mk].reshape((-1, 3, 1)), axis=0)
        norm3[i] = nsum(thigh_rot[mk].reshape((-1, 3, 1)) @ thigh_rot[mk].reshape((-1, 1, 3))
                        @ thigh_rot[mk].reshape((-1, 3, 1)), axis=0)
    norm1 /= n
    norm2 /= n
    norm3 /= n

    A = zeros((3, 3))
    b = zeros((3, 1))

    for i, mk in enumerate(thigh_rot.keys()):
        A += nsum(thigh_rot[mk].reshape((-1, 3, 1)) @ thigh_rot[mk].reshape((-1, 1, 3)), axis=0) / n
        A -= norm1[i] @ norm1[i].T

        b += norm3[i] - norm1[i] * norm2[i]

    A *= 2

    # solve for the initial guess of the joint center location in the pelvis cluster reference frame
    c = solve(A, b)

    # BIAS COMPENSATION
    delta_c = array([1, 1, 1])  # change in joint center coordinates
    while (delta_c > tol).any():
        sigma2 = zeros(len(thigh_rot.keys()))
        for i, mk in enumerate(thigh_rot.keys()):
            u = thigh_rot[mk].reshape((-1, 3, 1)) - c
            u2bar = 1 / u.shape[0] * nsum(u.transpose([0, 2, 1]) @ u, axis=0)
            sigma2[i] = 1 / (4 * u2bar * u.shape[0]) * nsum((nsum(u * u, axis=1) - u2bar) ** 2)

        sigma2avg = mean(sigma2)
        delta_b = zeros((3, 1))
        for mk in thigh_rot.keys():
            delta_b += 1 / thigh_rot[mk].shape[0] * nsum(thigh_rot[mk].reshape((-1, 3, 1)) - c, axis=0)

        delta_b *= 2 * sigma2avg

        cnew = solve(A, b - delta_b)

        delta_c = (cnew - c) / c
        c = cnew.copy()

    return c


def static(static_data, hip_center_data, window, marker_names='default'):
    """
    Process data from a static calibration trial to create the anatomical frames and create constant cluster to
    anatomical frame rotations.

    Parameters
    ----------
    static_data : tuple
        3-tuple of dictionaries. The first is the dictionary of pelvis data, the second is the dictionary of
        left-thigh data, and the third is the dictionary of right-thigh data during a static standing trial.
    hip_center_data : tuple
        3-tuple of dictionaries. The first is the dictionary of pelvis data, the second is the dictionary of
        left-thigh data, and the third is the dictionary of right-thigh data during a trial where the hip
        joint center can be calculated (enough rotation in axes is present to compute the joint center).
    window : int
        Number of samples to window over for finding the most still section of data. Suggested is 1 second worth of
        samples.
    marker_names : {'default', pymotion.omc.utility.MarkerNames}, optional
        Either 'default', which will use the default marker names, or a modified MarkerNames object, with marker names
        as used in the keys of pelvis_data, left_thigh_data, and right_thigh_data.

    Returns
    -------
    pelvis : tuple
        3-tuple of pelvis anatomical frame, rotation from world to pelvis cluster frame, and constant rotation from
        segment frame to cluster frame.
    left_thigh : tuple
        3-tuple of left thigh anatomical frame, rotation from world to the left thigh cluster frame, and constant
        rotation from segment frame to cluster frame.
    right_thigh : tuple
        3-tuple of right thigh anatomical frame, rotation from world to the right thigh cluster frame, and constant
        rotation from segment frame to cluster frame.
    """
    # extract the data from the tuples
    pelvis_data, left_thigh_data, right_thigh_data = static_data
    pelvis_jc_data, left_thigh_jc_data, right_thigh_jc_data = hip_center_data

    # separate the data from the names
    raw_data = tuple(pelvis_data[name] for name in pelvis_data.keys()) \
               + tuple(left_thigh_data[name] for name in left_thigh_data.keys()) \
               + tuple(right_thigh_data[name] for name in right_thigh_data.keys())
    raw_names = list(pelvis_data.keys()) + list(left_thigh_data.keys()) + list(right_thigh_data.keys())

    # find the most still period in the data, and the index of that window
    still_data, still_ind = common.find_most_still(raw_data, window, return_index=True)

    # associate still data with the names
    markers = dict()
    for data, name in zip(still_data, raw_names):
        markers[name] = data

    # set the marker names
    if marker_names == 'default':
        names = utility.MarkerNames()
    else:
        names = marker_names

    # -----------------------------------------------
    #              PELVIS
    # -----------------------------------------------
    # compute the pelvis origin, will be needed later
    pelvis_o = utility.compute_pelvis_origin(markers[names.left_asis], markers[names.right_asis])

    # create a matrix with columns as the x, y, z, axes of the anatomical frame
    # this is also the rotation matrix from pelvis frame to world frame
    pelvis_af = segmentFrames.pelvis(markers, use_cluster=False, marker_names=names)

    # create the cluster frame
    pelvis_R_c_w = utility.create_cluster_frame(markers, 'pelvis', marker_names=names)

    # create the constant segment to cluster rotation matrix
    pelvis_R_s_c = pelvis_R_c_w.T @ pelvis_af

    # -----------------------------------------------
    #              JOINT CENTERS
    # -----------------------------------------------
    # first compute the anatomical frame using the data provided for joint center computation
    pelvis_jc_af = segmentFrames.pelvis(pelvis_jc_data, use_cluster=False, marker_names=names)

    # compute the origin using the data provided for joint center computation
    pelvis_jc_o = utility.compute_pelvis_origin(pelvis_jc_data[names.left_asis], pelvis_jc_data[names.right_asis])

    # compute the joint center locations in the pelvis frame
    right_jc_p = compute_hip_center(pelvis_jc_data, right_thigh_jc_data, pelvis_jc_af.transpose([0, 2, 1]), pelvis_jc_o,
                                    marker_names=names)
    left_jc_p = compute_hip_center(pelvis_jc_data, left_thigh_jc_data, pelvis_jc_af.transpose([0, 2, 1]), pelvis_jc_o,
                                   marker_names=names)

    # -----------------------------------------------
    #              Right THIGH
    # -----------------------------------------------
    # first transform the left hip joint center coordinates back into the world frame
    right_hjc = pelvis_af @ right_jc_p.flatten() + pelvis_o

    # create the anatomical frame matrix, also the left thigh to world rotation matrix
    rthigh_af = segmentFrames.thigh(markers, 'right', use_cluster=False, hip_joint_center=right_hjc, marker_names=names)

    # compute the cluster orientation, also cluster to world
    rthigh_R_c_w = utility.create_cluster_frame(markers, 'right_thigh', names)

    # compute the constant segment to cluster rotation matrix
    rthigh_R_s_c = rthigh_R_c_w.T @ rthigh_af

    # -----------------------------------------------
    #              LEFT THIGH
    # -----------------------------------------------
    # first transform the left hip joint center coordinates back into the world frame
    left_hjc = pelvis_af @ left_jc_p.flatten() + pelvis_o

    # create the anatomical frame matrix, also the left thigh to world rotation matrix
    lthigh_af = segmentFrames.thigh(markers, 'left', use_cluster=False, hip_joint_center=left_hjc, marker_names=names)

    # compute the cluster orientation, also cluster to world rotation
    lthigh_R_c_w = utility.create_cluster_frame(markers, 'left_thigh', names)

    # compute the constant segment to cluster rotation matrix
    lthigh_R_s_c = lthigh_R_c_w.T @ lthigh_af

    return (pelvis_af, pelvis_R_c_w, pelvis_R_s_c), (lthigh_af, lthigh_R_c_w, lthigh_R_s_c), (rthigh_af, rthigh_R_c_w,
                                                                                              rthigh_R_s_c)





"""
Creation of anatomical segment frames from optical motion capture data

GNU GPL v3.0
Lukas Adamowicz
V0.1 - April 10, 2019
"""
from numpy import cross, stack
from numpy.linalg import norm

from pykinematics.omc import utility


def pelvis(marker_data, use_cluster=False, R_s_c=None, marker_names='default'):
    """
    Create the pelvis anatomical frame.

    Parameters
    ----------
    marker_data : dictionary
        Dictionary of marker position data. Keys correspond to marker names
    use_cluster : bool, optional
        Use the cluster to segment rotation to compute the anatomical frame. Default is False.
    R_s_c : {None, numpy.ndarray}, optional
        If use_cluster is False, R_s_c is ignored. If use_cluster is True, then a 3x3 rotation matrix must be provided
        that is the rotation from the segment to cluster frame for the pelvis.
    marker_names : {'default', pymotion.omc.utility.MarkerNames}, optional
        Either 'default', which will use the default marker names, or a modified MarkerNames object, with marker names
        as used in the keys of pelvis_data, left_thigh_data, and right_thigh_data.

    Returns
    -------
    pelvis_af : numpy.ndarray
        3x3 matrix representing the pelvis anatomical frame. Also is the rotation from world frame into the pelvis
        anatomical frame. Columns are the x, y, z axes of the anatomical frame.
    """
    # get marker names
    if marker_names == 'default':
        names = utility.MarkerNames
    else:
        names = marker_names

    if use_cluster:
        # first compute the cluster orientation
        R_c_w = utility.create_cluster_frame(marker_data, 'pelvis', names)

        # compute the anatomical frame matrix, which is the world to segment rotation matrix
        pelvis_af = R_c_w @ R_s_c

    else:
        # compute the pelvis origin
        origin = utility.compute_pelvis_origin(marker_data[names.left_asis], marker_data[names.right_asis])

        z = marker_data[names.right_asis] - marker_data[names.left_asis]
        z /= norm(z) if z.ndim == 1 else norm(z, axis=1, keepdims=True)

        mid_psis = (marker_data[names.right_psis] + marker_data[names.left_psis]) / 2

        x_tmp = origin - mid_psis

        y = cross(z, x_tmp)
        y /= norm(y) if y.ndim == 1 else norm(y, axis=1, keepdims=True)

        x = cross(y, z)
        x /= norm(x) if x.ndim == 1 else norm(x, axis=1, keepdims=True)

        # create a matrix with columns as the x, y, z, axes of the anatomical frame
        pelvis_af = stack((x, y, z), axis=1) if x.ndim == 1 else stack((x, y, z), axis=2)

    return pelvis_af


def thigh(marker_data, side, use_cluster=True, R_s_c=None, hip_joint_center=None, marker_names='default'):
    """
    Create the pelvis anatomical frame.

    Parameters
    ----------
    marker_data : dictionary
        Dictionary of thigh marker position data. Keys correspond to marker names
    side : {'left', 'right'}
        Thigh side.
    use_cluster : bool, optional
        Use the cluster to segment rotation to compute the anatomical frame. Default is True.
    R_s_c : {None, numpy.ndarray}, optional
        If use_cluster is False, R_s_c is ignored. If use_cluster is True, then a 3x3 rotation matrix must be provided
        that is the rotation from the segment to cluster frame for the pelvis.
    hip_joint_center : {None, numpy.ndarray}, optional
        If use_cluster is False, hip_joint_center must be the location of the hip joint center in the world frame. If
        use_cluster is True, then it is ignored.
    marker_names : {'default', pymotion.omc.utility.MarkerNames}, optional
        Either 'default', which will use the default marker names, or a modified MarkerNames object, with marker names
        as used in the keys of marker_data

    Returns
    -------
    thigh_af : numpy.ndarray
        3x3 matrix representing the thigh anatomical frame. Also is the rotation from world frame into the thigh
        anatomical frame. Columns are the x, y, z axes of the anatomical frame.
    """
    # get marker names
    if marker_names == 'default':
        names = utility.MarkerNames
    else:
        names = marker_names

    if use_cluster:
        # first compute the cluster orientation
        if side == 'right':
            segment_name = 'right_thigh'
        elif side == 'left':
            segment_name = 'left_thigh'
        else:
            raise ValueError('side must be either "left" or "right".')
        R_c_w = utility.create_cluster_frame(marker_data, segment_name, names)

        # compute the anatomical frame matrix, which is the segment to world rotation matrix
        thigh_af = R_c_w @ R_s_c

    else:
        # compute the midpoint of the epicondyles
        if side =='right':
            mid_ep = (marker_data[names.right_lep] + marker_data[names.right_mep]) / 2
        elif side == 'left':
            mid_ep = (marker_data[names.left_lep] + marker_data[names.left_mep]) / 2
        else:
            raise ValueError('side must be either "left" or "right".')

        # create the axes
        y = hip_joint_center - mid_ep
        y /= norm(y)

        if side == 'right':
            z_tmp = marker_data[names.right_lep] - marker_data[names.right_mep]
        elif side == 'left':
            z_tmp = marker_data[names.left_mep] - marker_data[names.left_lep]


        x = cross(y, z_tmp)
        x /= norm(x)

        z = cross(x, y)
        z /= norm(z)

        # create the anatomical frame matrix, also the world to left thigh rotation matrix
        thigh_af = stack((x, y, z), axis=1)

    return thigh_af




import pytest
from numpy import array


def pytest_addoption(parser):
    parser.addoption(
        "--no-integration", action="store_true", default=False, help="Don't run integration testing"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "integration: mark test as an integration test")


def pytest_collection_modifyitems(config, items):
    if not config.getoption('--no-integration'):
        return  # --no-integration wasn't given, run all tests
    skip_integration = pytest.mark.skip(reason="need --no-integration option to not run")
    for item in items:
        if "integration" in item.keywords:
            item.add_marker(skip_integration)


@pytest.fixture
def simple_x():
    return array([1, 9, 4, 2, 1, 3, 5, 4, 3, 5, 1, 3, 8, 4, 7, 6, 1, 2])


@pytest.fixture
def ma_window():
    return 5


@pytest.fixture
def simple_x_mov_stats():
    mn = array([0.0, 0.0, 0.0, 3.4, 3.8, 3.0, 3.0, 3.2, 4.0, 3.6, 3.2, 4.0, 4.2, 4.6, 5.6, 5.2, 0.0, 0.0])
    sd = array([0.0, 0.0, 0.0, 3.36154726, 3.1144823, 1.58113883, 1.58113883, 1.4832397, 1.0, 1.67332005, 1.4832397,
                2.64575131, 2.58843582, 2.88097206, 2.07364414, 2.77488739, 0.0, 0.0])
    return mn, sd



"""
General utility methods used in the calculation of joint angles from inertial sensors

GNU GPL v3.0
Lukas Adamowicz

V0.1 - March 8, 2019
"""
from numpy import array, zeros, argsort, dot, arccos, cross, cos, sin, real, allclose
from numpy.linalg import norm, eig


def calc_derivative(x, dt, order=2):
    """
    Calculate the 2nd or 4th order derivative of a sequence.

    Parameters
    ----------
    x : numpy.ndarray
        1 or 2D array to take the derivative of.
    dt : float
        Time difference for all points.  This cannot handle varying time differences.
    order : {4, 2}, optional
        Order of the derivative to calculate.  Default is 2nd order.

    Returns
    -------
    dx : numpy.ndarray
        Derivative of x.
    """
    dx = zeros(x.shape)

    if order == 4:
        dx[2:-2] = (x[:-4] - 8 * x[1:-3] + 8 * x[3:-1] + x[4:]) / (12 * dt)

        # edges
        dx[:2] = (-25 * x[:2] + 48 * x[1:3] - 36 * x[2:4] + 16 * x[3:5] - 3 * x[4:6]) / (12 * dt)
        dx[-2:] = (25 * x[-2:] - 48 * x[-3:-1] + 36 * x[-4:-2] - 16 * x[-5:-3] + 3 * x[-6:-4]) / (12 * dt)
    elif order == 2:
        dx[1:-1] = (x[2:] - x[:-2]) / (2 * dt)

        # edges
        dx[0] = (-3 * x[0] + 4 * x[1] - x[2]) / (2 * dt)
        dx[-1] = (3 * x[-1] - 4 * x[-2] + x[-3]) / (2 * dt)

    return dx


def quat_mult(q1, q2):
    """
    Multiply quaternions

    Parameters
    ----------
    q1 : numpy.ndarray
        1x4 array representing a quaternion
    q2 : numpy.ndarray
        1x4 array representing a quaternion

    Returns
    -------
    q : numpy.ndarray
        1x4 quaternion product of q1*q2
    """
    if q1.shape != (1, 4) and q1.shape != (4, 1) and q1.shape != (4,):
        raise ValueError('Quaternions contain 4 dimensions, q1 has more or less than 4 elements')
    if q2.shape != (1, 4) and q2.shape != (4, 1) and q2.shape != (4,):
        raise ValueError('Quaternions contain 4 dimensions, q2 has more or less than 4 elements')
    if q1.shape == (4, 1):
        q1 = q1.T

    Q = array([[q2[0], q2[1], q2[2], q2[3]],
               [-q2[1], q2[0], -q2[3], q2[2]],
               [-q2[2], q2[3], q2[0], -q2[1]],
               [-q2[3], -q2[2], q2[1], q2[0]]])

    return q1 @ Q


def quat_conj(q):
    """
    Compute the conjugate of a quaternion

    Parameters
    ----------
    q : numpy.ndarray
        Nx4 array of N quaternions to compute the conjugate of.

    Returns
    -------
    q_conj : numpy.ndarray
        Nx4 array of N quaternion conjugats of q.
    """
    return q * array([1, -1, -1, -1])


def quat_inv(q):
    """
    Invert a quaternion

    Parameters
    ----------
    q : numpy.ndarray
        1x4 array representing a quaternion

    Returns
    -------
    q_inv : numpy.ndarray
        1x4 array representing the inverse of q
    """
    # TODO change this, then remove the test that it doesn't work on arrays of quaternions
    if q.size != 4:
        raise ValueError('Not currently implemented for arrays of quaternions.')
    q_conj = q * array([1, -1, -1, -1])
    return q_conj / sum(q ** 2)


def quat2matrix(q):
    """
    Transform quaternion to rotation matrix

    Parameters
    ----------
    q : numpy.ndarray
        Quaternion

    Returns
    -------
    R : numpy.ndarray
        Rotation matrix
    """
    if q.ndim == 1:
        s = norm(q)
        R = array([[1 - 2 * s * (q[2] ** 2 + q[3] ** 2), 2 * s * (q[1] * q[2] - q[3] * q[0]),
                    2 * s * (q[1] * q[3] + q[2] * q[0])],
                   [2 * s * (q[1] * q[2] + q[3] * q[0]), 1 - 2 * s * (q[1] ** 2 + q[3] ** 2),
                    2 * s * (q[2] * q[3] - q[1] * q[0])],
                   [2 * s * (q[1] * q[3] - q[2] * q[0]), 2 * s * (q[2] * q[3] + q[1] * q[0]),
                    1 - 2 * s * (q[1] ** 2 + q[2] ** 2)]])
    elif q.ndim == 2:
        s = norm(q, axis=1)
        R = array([[1 - 2 * s * (q[:, 2]**2 + q[:, 3]**2), 2 * s * (q[:, 1] * q[:, 2] - q[:, 3] * q[:, 0]),
                    2 * s * (q[:, 1] * q[:, 3] + q[:, 2] * q[:, 0])],
                   [2 * s * (q[:, 1] * q[:, 2] + q[:, 3] * q[:, 0]), 1 - 2 * s * (q[:, 1]**2 + q[:, 3]**2),
                    2 * s * (q[:, 2] * q[:, 3] - q[:, 1] * q[:, 0])],
                   [2 * s * (q[:, 1] * q[:, 3] - q[:, 2] * q[:, 0]), 2 * s * (q[:, 2] * q[:, 3] + q[:, 1] * q[:, 0]),
                    1 - 2 * s * (q[:, 1]**2 + q[:, 2]**2)]])
        R = R.transpose([2, 0, 1])
    return R


def quat_mean(q):
    """
    Calculate the mean of an array of quaternions

    Parameters
    ----------
    q : numpy.ndarray
        Nx4 array of quaternions

    Returns
    -------
    q_mean : numpy.array
        Mean quaternion
    """
    M = q.T @ q

    vals, vecs = eig(M)  # Eigenvalues and vectors of M
    sort_ind = argsort(vals)  # get the indices of the eigenvalues sorted

    q_mean = real(vecs[:, sort_ind[-1]])

    # ensure no discontinuities
    if q_mean[0] < 0:
        q_mean *= -1

    return q_mean


def vec2quat(v1, v2):
    """
    Find the rotation quaternion between two vectors. Rotate v1 onto v2

    Parameters
    ----------
    v1 : numpy.ndarray
        Vector 1
    v2 : numpy.ndarray
        Vector 2

    Returns
    -------
    q : numpy.ndarray
        Quaternion representing the rotation from v1 to v2
    """
    if allclose(v1, v2):
        return array([1, 0, 0, 0])
    else:
        angle = arccos(dot(v1.flatten(), v2.flatten()) / (norm(v1) * norm(v2)))

        # Rotation axis is always normal to two vectors
        axis = cross(v1.flatten(), v2.flatten())
        axis = axis / norm(axis)  # normalize

        q = zeros(4)
        q[0] = cos(angle / 2)
        q[1:] = axis * sin(angle / 2)
        q /= norm(q)

        return q



from pykinematics.imu.lib import angles
from pykinematics.imu.lib import calibration
from pykinematics.imu.lib import joints
from pykinematics.imu import optimize
from pykinematics.imu import orientation
from pykinematics.imu import utility

__all__ = ['optimize', 'orientation', 'angles', 'calibration', 'joints']



"""
Optimization filters

GNU GPL v3.0
Lukas Adamowicz

V0.1 - March 8, 2019
"""
from numpy import zeros, sqrt
from numpy.linalg import inv, cholesky, pinv

__all__ = ['UnscentedKalmanFilter', 'GaussNewton']


class UnscentedKalmanFilter:
    def __init__(self, x0, P0, F, H, Q, R):
        """
        Parameters
        ----------
        x0 : matrix_like, array_like
            Initial guess for state column vector (Nx1).
        P0 : matrix_like, array_like
            Initial guess for state covariance.  Must be NxN size.
        F : callable
            Function to time-step state vector.  x(t+dt) = F[x(t)]
        H : callable
            Function to estimate measured vector.  z(t) = H[x(t);z(t)]
        Q : matrix_like,array_like,float,optional
            State covariance.  Must be None, NxN matrix/array, or float.  If None,
            defaults to 1.5*I where I is the NxN identity matrix
        R : matrix_like,array_like,float,str,optional
            Process covariance.  Must be None, NxN matrix/array, or float.
            If None, defaults to 1.
        """

        if x0.ndim == 1:
            self.x = x0.reshape((-1, 1))
        else:
            self.x = x0

        self.P = P0
        self.F = F
        self.H = H
        self.Q = Q
        self.R = R

    def run(self, z, f_kwargs={}, h_kwargs={}, sigma_kwargs={}):
        """
        Run Unscented Kalman filter for 1 step

        Parameters
        ----------
        z : array_like
            mx1 array of measured data
        f_kwargs : dict,optional
            Any additional keyword arguments to be passed to state update function
            F.  Optional, defaults to no arguments passed.
        h_kwargs : dict,optional
            Any additional keyword arguments to be passed to measured input estimation
            function H.  Optional, defaults to no arguments passed.
        sigma_kwargs : dict, optional
            Sigma point computation key-word arguments.

        Returns
        ----------------
        x : array_like
            Estimated state variable
        """
        # make sure that z is 2D for proper broadcasting
        if z.ndim == 1:
            z.resize((z.size, 1))

        # compute the sigma points from the state and state covariance
        Xi, W = UnscentedKalmanFilter.sigma_points(self.x, self.P, **sigma_kwargs)

        # update the sigma points
        fXi = self.F(Xi, **f_kwargs)

        # transform sigma points back to normal space
        xp, Px = UnscentedKalmanFilter.unscented_transform(fXi, W, self.Q)

        # get the measurement prediction from the sigma points
        hXi = self.H(fXi, **h_kwargs)

        # transform the measurement prediction to normal space
        zp, Pz = UnscentedKalmanFilter.unscented_transform(hXi, W, self.R)

        Pxz = zeros((self.x.shape[0], z.shape[0]))
        for k in range(2 * self.x.shape[0] + 1):
            Pxz += W[k] * (fXi[:, k].reshape((-1, 1)) - xp) @ (hXi[:, k].reshape((-1, 1)) - zp).T

        # calculate the Kalman gain
        K = Pxz @ inv(Pz)

        # get the best estimate of the state and state covariance
        self.x = xp + K @ (z - zp)
        self.P = Px - K @ Pz @ K.T

        return self.x

    @staticmethod
    def sigma_points(x, P, alpha=0.001, kappa=3.0, beta=2):
        """
        Calculate Sigma Points

        Parameters
        ----------
        x : array_like
            Vector to calculate sigma points for.
        P : array_like
            Covariance array used in sigma point calculation.
        alpha : float, optional
            Alpha value used in sigma point calculation.  Defaults to 0.001.
        kappa : float, optional
            Kappa value used in sigma point calculation.  Defaults to 0.0.
        beta : float, optional
            Beta value, based on the distribution of whats being estimated.

        Returns
        -------
        xi : array
            Sigma points
        W : array
            Weights
        """
        n = x.size
        Xi = zeros((n, 2 * n + 1))
        W = zeros((2 * n + 1, 1))

        # wiki also has different weights for the state and covariance
        # lambda in wiki, kappa in book
        y = alpha ** 2 * (n + kappa) - n

        Xi[:, 0] = x.flatten().copy()
        W[0] = y / (n + y)

        U = cholesky((n + y) * P)

        for i in range(n):
            Xi[:, i + 1] = x.flatten() + U[i, :]
            Xi[:, i + n + 1] = x.flatten() - U[i, :]

        W[1:] = 1 / (2 * (n + y))

        return Xi, W

    @staticmethod
    def unscented_transform(Xi, W, noise_cov=0.0):
        """
        Unscented transform

        Parameters
        ----------
        Xi : array_like
            Sigma Points
        W : array_like
            Weights for sigma points
        noise_cov : float,optional
            Noise covariance.  Optional, defaults to 0.0

        Returns
        -------
        xm : array_like
            Mean of input sigma points from unscented transform
        xcov : array_like
            Covariance of input sigma points from unscented transform
        """
        n, kmax = Xi.shape

        x = Xi @ W

        P = zeros((n, n))
        for k in range(kmax):
            P += W[k] * (Xi[:, k].reshape((-1, 1)) - x) @ (Xi[:, k].reshape((-1, 1)) - x).T

        P += noise_cov

        return x, P


class GaussNewton:
    def __init__(self, x0, tol=0.001, f_tol=0.001, max_iters=300, verbose=False):
        """
        Gauss-Newton optimization function

        Parameters
        ----------
        x0 : numpy.ndarray
            Mx1 array of initial guess for x
        tol : float, optional
            Tolerance for the change in x per iteration, as a percentage of x.  Default is 0.001.
        f_tol : float, optional
            Tolerance for the change in the residual value per iteration, as a percentage of previous residual.
            Default is 0.001
        max_iters : int, optional
            Maximum number of iterations.  Default is 300
        verbose : bool, optional
            Display convergence messages on the first 10 iterations, and then on iterations that are perfect squares.
            Default is False
        """
        self.tol = tol
        self.ftol = f_tol
        self.max_iters = max_iters
        self.verbose = verbose

        if x0.ndim == 1:
            self.x = x0.reshape((-1, 1)).copy()
        else:
            self.x = x0.copy()

    def fit(self, f, J, f_args=(), f_kwargs={}, j_args=(), j_kwargs={}):
        """
        Perform the optimization on x.

        Parameters
        ----------
        f : callable
            Function to calculate the Nx1 array of residuals.  First argument is x, the parameters being optimized.
        J : callable
            Function to calculate the NxM Jacobian matrix.  First argument is x, the parameters being optimized.
        f_args : tuple, optional
            Additional arguments to pass to the residual calculation function.
        f_kwargs : dict, optional
            Additional key-word arguments to pass to the residual calculation function.
        j_args : tuple, optional
            Additional arguments to pass to the Jacobian calculation function.
        j_kwargs : dict, optional
            Additional key-word arguments to pass to the residual calculation function.

        Returns
        -------
        x : numpy.ndarray
            Optimized result
        i : int
            Number of iterations to converge on solution
        fval : float
            Sum of squared errors at final solution
        """

        # initialize stopping parameters
        dx = self.x.copy()
        pc = (abs(dx / self.x) > self.tol).any()  # percent change in x is larger than tolerance
        fval = 1e12
        fc = 1 > self.ftol  # percent change in sum of squared error is larger than error tolerance
        i =0

        while pc and fc and i < self.max_iters:
            e = f(self.x, *f_args, **f_kwargs)  # calculate the residual vector
            j = J(self.x, *j_args, **j_kwargs)  # calculate the jacobian

            j_inv = pinv(j)  # calculate the moore-penrose inverse of J

            dx = j_inv @ e  # calculate the update to x

            # calculate percent change in x parameter
            pc = (abs(dx / self.x) > self.tol).any()

            # update x
            self.x -= dx

            # calculate percent change in SSE, and SSE
            fc = abs(sum(e ** 2) - fval) / fval
            fval = sum(e ** 2)

            # print message if verbose is True
            if self.verbose:
                if sqrt(i) % 1 == 0 or i < 11:
                    print(f'{i:5d}  SSE per sample: {fval/e.shape[0]:10.3f}')

            # increment i
            i += 1

        return self.x, i, fval



"""
Methods for calculating joint angles using previously computed joint parameters and calibrated frames

GNU GPL v3.0
Lukas Adamowicz

V0.1 - March 8, 2019
"""
from numpy import cross, abs as nabs, arctan2 as atan2, sum, arccos, pi, stack
from numpy.linalg import norm
from scipy.integrate import cumtrapz


__all__ = ['hip_from_frames']


def hip_from_frames(pelvis_AF, thigh_AF, R, side, zero_angles=False):
    """
    Compute the hip joint angles from the segment fixed axes computed during the calibration.

    Parameters
    ----------
    pelvis_AF : tuple
        Tuple of the x, y, and z axes of the pelvis anatomical frame.
    thigh_AF : tuple
        Tuple of the x, y, and z axes of the thigh anatomical frame.
    R : numpy.ndarray
        Nx3x3 array of rotation matrices from the thigh sensor frame to the pelvis sensor frame.
    side : {'left', 'right'}
        Side the angles are being computed for.
    zero_angles : bool, optional
        Remove any offset from zero at the start of the angles. Default is False.

    Returns
    -------
    angles : numpy.ndarray
        Nx3 array of hip angles in degrees, with the first column being Flexion - Extension,
        second column being Ad / Abduction, and the third column being Internal - External Rotation.
    """
    # get pelvis anatomical frame
    X = pelvis_AF[0]
    Y = pelvis_AF[1]
    Z = pelvis_AF[2]

    # get the thigh anatomical frame and rotate into pelvis frame
    x = R @ thigh_AF[0]
    y = R @ thigh_AF[1]
    z = R @ thigh_AF[2]

    # form the joint axes
    e1 = Z.copy()
    e3 = y.copy()

    e2 = cross(e3, e1)
    e2 /= norm(e2, axis=1, keepdims=True)

    # compute angles
    sgn = sum(cross(X, e2) * Z, axis=1)
    sgn /= nabs(sgn)
    fe = atan2(sgn * norm(cross(X, e2), axis=1), sum(X * e2, axis=1))

    if side == 'right':
        # ad / abduction calculation
        aa = -pi / 2 + arccos(sum(e1 * e3, axis=1))

        # internal - external rotation sign calculation
        sgn = sum(cross(x, e2) * -y, axis=1)
    elif side == 'left':
        aa = pi / 2 - arccos(sum(e1 * e3, axis=1))
        sgn = sum(cross(x, e2) * y, axis=1)
    else:
        raise ValueError("side must be either 'left' or 'right'.")

    sgn /= nabs(sgn)
    ier = atan2(sgn * norm(cross(x, e2), axis=1), sum(x * e2, axis=1))

    angles = stack((fe, aa, ier), axis=1) * 180 / pi
    if zero_angles:
        angles -= angles[0, :]

    return angles


def hip_from_gyr(pelvis_w, thigh_w, pelvis_axis, thigh_axis, R, side):
    """
    Compute hip angles by integrating the difference between the two sensors angular velocities about the axis
    of rotation.  Typically not very noisy on higher frequencies, but on lower frequencies where it exhibits drift
    due to integration being used.

    Parameters
    ----------
    pelvis_w : numpy.ndarray
        Nx3 array of angular velocity vectors in the pelvis sensor frame.
    thigh_w : numpy.ndarray
        Nx3 array of angular velocity vectors in the thigh sensor frame.
    pelvis_axis : numpy.ndarray
        Pelvis fixed axis.
    thigh_axis : numpy.ndarray
        Thigh fixed axis.
    R : numpy.ndarray
        Nx3x3 array of rotation matrices from the thigh sensor frame to the pelvis sensor frame.
    side : {'left', 'right'}
        Side angles are being computed for.

    Returns
    -------
    angles : numpy.ndarray
        Nx3 array of joint angles for the trial.  Columns are flexion - extension, ad / abduction, and
        internal - external rotation respectively.
    """
    # get the joint rotation axes in the pelvis frame
    pelvis_e1 = pelvis_axis
    pelvis_e3 = R @ thigh_axis

    pelvis_e2 = cross(pelvis_e3, pelvis_e1)
    pelvis_e2 /= norm(pelvis_e2, axis=1, keepdims=True)

    # get the joint rotation axes in the thigh frame
    thigh_e1 = R.transpose([0, 2, 1]) @ pelvis_axis
    thigh_e3 = thigh_axis

    thigh_e2 = cross(thigh_e3, thigh_e1)
    thigh_e2 /= norm(thigh_e2, axis=1, keepdims=True)

    # compute the differences between angular rates around their respective rotation axes
    # TODO needs to be corrected still
    if side == 'left':
        fe_int = sum(pelvis_w * pelvis_e1, axis=1) - sum(thigh_w * thigh_e1, axis=1)
        aa_int = -sum(pelvis_w * pelvis_e2, axis=1) + sum(thigh_w * thigh_e2, axis=1)
        ier_int = sum(pelvis_w * pelvis_e3, axis=1) - sum(thigh_w * thigh_e3, axis=1)
    elif side == 'right':
        fe_int = sum(pelvis_w * pelvis_e1, axis=1) - sum(thigh_w * thigh_e1, axis=1)
        aa_int = sum(pelvis_w * pelvis_e2, axis=1) - sum(thigh_w * thigh_e2, axis=1)
        ier_int = -sum(pelvis_w * pelvis_e3, axis=1) + sum(thigh_w * thigh_e3, axis=1)
    else:
        raise ValueError("side must be 'left' or 'right'.")

    # integrate the differences in angular velocities to get angles.
    fe = cumtrapz(fe_int, dx=1 / 128, initial=0)
    aa = cumtrapz(aa_int, dx=1 / 128, initial=0)
    ier = cumtrapz(ier_int, dx=1 / 128, initial=0)

    angles = stack((fe, aa, ier), axis=1) * 180 / pi
    return angles



"""
Methods for calculating joint parameters, such as joint centers and axes.

GNU GPL v3.0
Lukas Adamowicz

V0.1 - March 8, 2019
"""
from numpy import array, zeros, logical_and, abs as nabs, concatenate, cross, std, sign, argmax, sqrt
from numpy.linalg import lstsq, norm
from scipy.optimize import least_squares


__all__ = ['Center', 'KneeAxis', 'correct_knee', 'fixed_axis']


class Center:
    def __init__(self, g=9.81, method='SAC', mask_input=True, min_samples=1000, mask_data='acc', opt_kwargs=None):
        """
        Estimation of joint centers using kinematic constraints.

        Parameters
        ----------
        g : float, optional
            Local value of gravitational acceleration. Default is 9.81 m/s^2.
        method : {'SAC', 'SSFC'}, optional
            Method to use for the computation of the joint center. Default is SAC. See Crabolu et al. for more details.
            Method 'SAC' requires knowing the rotation between the sensors.
        mask_input : bool, optional
            Mask the input to only use the highest acceleration samples. Default is True
        min_samples : int, optional
            Minimum number of samples to use. Default is 1000.
        mask_data : {'acc', 'gyr'}
            Data to use for masking. Default is acceleration.
        opt_kwargs : dict, optional
            Optimization key-word arguments. SAC uses numpy.linalg.lstsq.
            SSFC and SSFCv use scipy.optimize.least_squares.

        References
        ----------
        Crabolu et al. "In vivo estimation of the shoulder joint center of rotation using magneto-inertial sensors:
        MRI-based accuracy and repeatability assessment." *BioMedical Engineering Online*. 2017.
        """
        self.g = g
        if method in ['SAC', 'SSFC']:
            self.method = method
        else:
            raise ValueError("Method must be either 'SAC' or 'SSFC'.")
        self.mask_input = mask_input
        self.min_samples = min_samples
        if mask_data in ['acc', 'gyr']:
            self.mask_data = mask_data
        else:
            raise ValueError("Masking data must be either 'acc' or 'gyr'.")
        if opt_kwargs is not None:
            self.opt_kwargs = opt_kwargs
        else:
            self.opt_kwargs = {}

    def compute(self, prox_a, dist_a, prox_w, dist_w, prox_wd, dist_wd, R_dist_prox):
        """
        Perform the computation of the joint center to sensor vectors.

        Parameters
        ----------
        prox_a : numpy.ndarray
            Nx3 array of accelerations measured by the joint proximal sensor.
        dist_a : numpy.ndarray
            Nx3 array of accelerations measured by the joint distal sensor.
        prox_w : numpy.ndarray
            Nx3 array of angular velocities measured by the joint proximal sensor.
        dist_w : numpy.ndarray
            Nx3 array of angular velocities measured by the joint distal sensor.
        prox_wd : numpy.ndarray
            Nx3 array of angular accelerations measured by the joint proximal sensor.
        dist_wd : numpy.ndarray
            Nx3 array of angular accelerations measured by the joint distal sensor.
        R_dist_prox : numpy.ndarray
            Nx3x3 array of rotations from the distal sensor frame to the proximal sensor frame. Ignored if method is
            'SSFC'.

        Returns
        -------
        prox_r : numpy.ndarray
            Joint center to proximal sensor vector.
        dist_r : numpy.ndarray
            Joint center to distal sensor vector.
        residual : float
            Residual value per sample used from the joint center optimization
        """
        if self.method == 'SAC':
            if self.mask_input:
                if self.mask_data == 'acc':
                    prox_data = norm(prox_a, axis=1) - self.g
                    dist_data = norm(dist_a, axis=1) - self.g
                    thresh = 1.0
                elif self.mask_data == 'gyr':
                    prox_data = norm(prox_w, axis=1)
                    dist_data = norm(dist_w, axis=1)
                    thresh = 2.0

                mask = zeros(prox_data.shape, dtype=bool)

                while mask.sum() < self.min_samples:
                    mask = logical_and(nabs(prox_data) > thresh, nabs(dist_data) > thresh)

                    thresh -= 0.05
                    if thresh < 0.09:
                        raise ValueError('Not enough samples or samples with high motion in the trial provided.  '
                                         'Use another trial')
            else:
                mask = zeros(prox_a.shape[0], dtype=bool)
                mask[:] = True

            # create the skew symmetric matrix products
            prox_K = array([[-prox_w[mask, 1] ** 2 - prox_w[mask, 2] ** 2,
                             prox_w[mask, 0] * prox_w[mask, 1] - prox_wd[mask, 2],
                             prox_wd[mask, 1] + prox_w[mask, 0] * prox_w[mask, 2]],
                            [prox_wd[mask, 2] + prox_w[mask, 0] * prox_w[mask, 1],
                             -prox_w[mask, 0] ** 2 - prox_w[mask, 2] ** 2,
                             prox_w[mask, 1] * prox_w[mask, 2] - prox_wd[mask, 0]],
                            [prox_w[mask, 0] * prox_w[mask, 2] - prox_wd[mask, 1],
                             prox_wd[mask, 0] + prox_w[mask, 1] * prox_w[mask, 2],
                             -prox_w[mask, 0] ** 2 - prox_w[mask, 1] ** 2]]).transpose([2, 0, 1])

            dist_K = array([[-dist_w[mask, 1] ** 2 - dist_w[mask, 2] ** 2,
                             dist_w[mask, 0] * dist_w[mask, 1] - dist_wd[mask, 2],
                             dist_wd[mask, 1] + dist_w[mask, 0] * dist_w[mask, 2]],
                            [dist_wd[mask, 2] + dist_w[mask, 0] * dist_w[mask, 1],
                             -dist_w[mask, 0] ** 2 - dist_w[mask, 2] ** 2,
                             dist_w[mask, 1] * dist_w[mask, 2] - dist_wd[mask, 0]],
                            [dist_w[mask, 0] * dist_w[mask, 2] - dist_wd[mask, 1],
                             dist_wd[mask, 0] + dist_w[mask, 1] * dist_w[mask, 2],
                             -dist_w[mask, 0] ** 2 - dist_w[mask, 1] ** 2]]).transpose([2, 0, 1])

            # create the oversized A and b matrices
            A = concatenate((prox_K, -R_dist_prox[mask] @ dist_K), axis=2).reshape((-1, 6))
            b = (prox_a[mask].reshape((-1, 3, 1))
                 - R_dist_prox[mask] @ dist_a[mask].reshape((-1, 3, 1))).reshape((-1, 1))

            # solve the linear least squares problem
            r, residual, _, _ = lstsq(A, b, rcond=None)
            r.resize((6,))
            residual = residual[0]

        elif self.method == 'SSFC':
            r_init = zeros((6,))

            if self.mask_input:
                if self.mask_data == 'acc':
                    prox_data = norm(prox_a, axis=1) - self.g
                    dist_data = norm(dist_a, axis=1) - self.g
                    thresh = 1.0
                elif self.mask_data == 'gyr':
                    prox_data = norm(prox_w, axis=1)
                    dist_data = norm(dist_w, axis=1)
                    thresh = 2.0

                mask = zeros(prox_data.shape, dtype=bool)

                while mask.sum() < self.min_samples:
                    mask = logical_and(nabs(prox_data) > thresh, nabs(dist_data) > thresh)

                    thresh -= 0.05
                    if thresh < 0.09:
                        raise ValueError('Not enough samples or samples with high motion in the trial provided.  '
                                         'Use another trial')
            else:
                mask = zeros(prox_a.shape[0], dtype=bool)
                mask[:] = True

            # create the arguments to be passed to both the residual and jacobian calculation functions
            args = (prox_a[mask], dist_a[mask], prox_w[mask], dist_w[mask], prox_wd[mask], dist_wd[mask])

            sol = least_squares(Center._compute_distance_residuals, r_init.flatten(), args=args, **self.opt_kwargs)
            r = sol.x
            residual = sol.cost

        return r[:3], r[3:], residual / mask.sum()

    @staticmethod
    def _compute_distance_residuals(r, a1, a2, w1, w2, wd1, wd2):
        """
            Compute the residuals for the given joint center locations for proximal and distal inertial data

            Parameters
            ----------
            r : numpy.ndarray
                6x1 array of joint center locations.  First three values are proximal location guess, last three values
                are distal location guess.
            a1 : numpy.ndarray
                Nx3 array of accelerations from the proximal sensor.
            a2 : numpy.ndarray
                Nx3 array of accelerations from the distal sensor.
            w1 : numpy.ndarray
                Nx3 array of angular velocities from the proximal sensor.
            w2 : numpy.ndarray
                Nx3 array of angular velocities from the distal sensor.
            wd1 : numpy.ndarray
                Nx3 array of angular accelerations from the proximal sensor.
            wd2 : numpy.ndarray
                Nx3 array of angular accelerations from the distal sensor.

            Returns
            -------
            e : numpy.ndarray
                Nx1 array of residuals for the given joint center location guess.
            """
        r1 = r[:3]
        r2 = r[3:]

        at1 = a1 - cross(w1, cross(w1, r1, axisb=0)) - cross(wd1, r1, axisb=0)
        at2 = a2 - cross(w2, cross(w2, r2, axisb=0)) - cross(wd2, r2, axisb=0)

        return norm(at1, axis=1) - norm(at2, axis=1)


class KneeAxis:
    def __init__(self, mask_input=True, min_samples=1500, opt_kwargs=None):
        """
        Estimation of the flexion/extension knee axis using kinematic constraints.

        Parameters
        ----------
        mask_input : bool, optional
            Mask the input to only use samples with enough angular velocity to give a good estimate.  Default is True.
        min_samples : int, optional
            Minimum number of samples to use in the optimization.  Default is 1500.
        opt_kwargs : dict, optional
            Optimization key-word arguments.  See scipy.optimize.least_squares.

        References
        ----------
        Seel et al. "IMU-Based Joint Angle Measurement for Gait Analysis." *Sensors*. 2014
        Seel et al. "Joint axis and position estimation from inertial measurement data by exploiting kinematic
        constraints." *2012 IEEE International Conference on Control Applications*. 2012
        """
        self.mask_input = mask_input
        self.min_samples = min_samples
        if opt_kwargs is not None:
            self.opt_kwargs = opt_kwargs
        else:
            self.opt_kwargs = {}

    def compute(self, thigh_w, shank_w):
        """
        Compute the knee axis using the given angular velocities.

        Parameters
        ----------
        thigh_w : numpy.ndarray
            Nx3 array of angular velocities measured by the thigh sensor.
        shank_w : numpy.ndarray
            Nx3 array of angular velocities measured by the shank sensor.

        Returns
        -------
        thigh_j : numpy.ndarray
            Vector of the joint rotation axis in the thigh sensor frame.
        shank_j : numpy.ndarray
            Vector of the joint rotation axis in the shank sensor frame.
        """
        j_init = (zeros(6) + 1) / sqrt(3)

        if self.mask_input:
            thigh_wn = norm(thigh_w, axis=1)
            shank_wn = norm(shank_w, axis=1)

            factor = 1.0
            mask = zeros(thigh_wn.size, dtype=bool)
            while mask.sum() < self.min_samples:
                mask = (thigh_wn > (factor * std(thigh_wn))) & (shank_wn > (factor * std(shank_wn)))
                factor -= 0.1
                if factor < 0.15:
                    raise ValueError('Not enough samples to mask and still estimate the joint axis.  Consider not '
                                     'masking, or use a trial with more samples')
        else:
            mask = array([True] * thigh_w.shape[0])

            # arguments for the solver
        args = (thigh_w[mask], shank_w[mask])
        # solve and normalize the result
        sol = least_squares(KneeAxis._compute_axis_residuals, j_init.flatten(), args=args, **self.opt_kwargs)
        thigh_j = sol.x[:3] / norm(sol.x[:3])
        shank_j = sol.x[3:] / norm(sol.x[3:])

        return thigh_j, shank_j

    @staticmethod
    def _compute_axis_residuals(j, thigh_w, shank_w):
        """
        Compute the residuals using the estimate of the axis.

        Parameters
        ----------
        j : numpy.ndarray
            Estimates of thigh and shank axis stacked into one vector
        thigh_w : numpy.ndarray
            Nx3 array of angular velocities measured by the thigh sensor.
        shank_w : numpy.ndarray
            Nx3 array of angular velocities measured by the shank sensor.

        Returns
        -------
        e : numpy.ndarray
            N length array of residuals.
        """
        j1 = j[:3] / norm(j[:3])
        j2 = j[3:] / norm(j[3:])

        wp1 = cross(thigh_w, j1)
        wp2 = cross(shank_w, j2)

        return norm(wp1, axis=1) - norm(wp2, axis=1)


def correct_knee(thigh_w, shank_w, thigh_r, shank_r, R_shank_thigh, knee_axis_kwargs=None):
    """
    Correct the knee position based on the computed knee axis.

    Parameters
    ----------
    thigh_w : numpy.ndarray
        (N, 3) array of angular velocities measured by the thigh sensor.
    shank_w : numpy.ndarray
        (N, 3) array of angular velocities measured by the shank sensor.
    thigh_r : numpy.ndarray
        Initial knee joint center to thigh sensor vector.
    shank_r : numpy.ndarray
        Initial knee joint center to shank sensor vector.
    R_shank_thigh : numpy.ndarray
        (3, 3) Rotation matrix from shank to thigh sensors.
    knee_axis_kwargs : dict, optional
        Knee axis computation key-word arguments. See KneeAxis.

    Returns
    -------
    thigh_r_corr : numpy.ndarray
        Corrected knee joint center to thigh sensor vector.
    shank_r_corr : numpy.ndarray
        Corrected knee joint center to shank sensor vector.
    """
    if knee_axis_kwargs is None:
        knee_axis_kwargs = {}
    # compute the knee axis
    ka = KneeAxis(**knee_axis_kwargs)
    thigh_j, shank_j = ka.compute(thigh_w, shank_w)

    # check the sign of the major component of the axes when rotated into the same frame
    shank_j_thigh = R_shank_thigh @ shank_j
    if sign(shank_j_thigh[argmax(nabs(shank_j_thigh))]) != sign(thigh_j[argmax(nabs(thigh_j))]):
        shank_j *= -1

    # compute the corrections for the joint centers
    tmp = (sum(thigh_r * thigh_j) + sum(shank_r * shank_j)) / 2
    thigh_r_corr = thigh_r - thigh_j * tmp
    shank_r_corr = shank_r - shank_j * tmp

    return thigh_r_corr, shank_r_corr


def fixed_axis(center1, center2, center_to_sensor=True):
    """
    Compute the fixed axis of a segment based on computed joint centers.

    Parameters
    ----------
    center1 : numpy.ndarray
        Location of the first joint center. This will be the "origin" of the axis.
    center2 : numpy.ndarray
        Location of the second joint center. This will be the "end" of the axis.
    center_to_sensor : bool, optional
        If the vectors provided are joint center to sensor (opposite is sensor to joint center). If True, then the
        axes are created in the opposite way as expected (eg right pointing pelvis axis would be left hip joint center
        minus the right hip joint center). Default is True.

    Returns
    -------
    axis : numpy.ndarray
        Fixed axis based on joint centers, from center1 to center2.
    """

    if center_to_sensor:
        axis = center1 - center2
    else:
        axis = center2 - center1

    # normalize
    axis /= norm(axis)
    return axis




"""
Methods for calibration of imu sensors for use in joint angle calculation

GNU GPL v3.0
Lukas Adamowicz

V0.1 - March 8, 2019
"""
from numpy import mean, ceil, cross
from numpy.linalg import norm

from pykinematics.imu import utility
from pykinematics import common


__all__ = ['get_acc_scale', 'static']


def get_acc_scale(acc, gravity=9.81):
    """
    Get the acceleration scale factor so that the mean of the acceleration during the static trial is equal between
    sensors.

    Parameters
    ----------
    acc : numpy.ndarray
        Nx3 array of accelerations during a static calibration trial.
    gravity : float, optional
        Value of local gravitational acceleration. Default is 9.81m/s^2

    Returns
    -------
    scale : float
        Scale factor to multiply acceleration by
    """
    mean_acc = mean(norm(acc, axis=1))

    return gravity / mean_acc


def static(lt_p_q, rt_p_q, pelvis_axis, l_thigh_axis, r_thigh_axis, pelvis_w, l_thigh_w, r_thigh_w, fs, window=1.0):
    """
    Process a static standing data to create the anatomical axes for the pelvis and thighs.

    Parameters
    ----------
    lt_p_q : numpy.ndarray
        Nx4 array of quaternions representing the rotation from the left thigh to the pelvis.
    rt_p_q : numpy.ndarray
        Nx4 array of quaternions representing the rotation from the right thigh to the pelvis.
    pelvis_axis : numpy.ndarray
        Pelvis fixed axis.
    l_thigh_axis : numpy.ndarray
        Left thigh fixed axis.
    r_thigh_axis : numpy.ndarray
        Right thigh fixed axis.
    pelvis_w : numpy.ndarray
        Nx3 array of angular velocities measured by the pelvis sensor.
    l_thigh_w : numpy.ndarray
        Nx3 array of angular velocities measured by the left thigh sensor.
    r_thigh_w : numpy.ndarray
        Nx3 array of angular velocities measured by the right thigh sensor.
    fs : float
        Sampling frequency of the sensors.
    window : float, optional
        Length of time for the most still period. Default is 1.0s.

    Returns
    -------
    pelvis_AF : tuple
        Tuple of the x, y, and z axes of the pelvis anatomical frame.
    l_thigh_AF : tuple
        Tuple of the x, y, and z axes of the left thigh anatomical frame.
    r_thigh_AF : tuple
        Tuple of the x, y, and z axes of the right thigh anatomical frame.
    """

    # find the most still period in the angular velocity data of the pelvis and thighs
    _, ind = common.find_most_still((pelvis_w, l_thigh_w, r_thigh_w), int(window * fs), return_index=True)

    pad = int(ceil(window * fs / 2))

    # compute the rotations during the most still time
    if (ind - pad) < 0:
        l_q_mean = utility.quat_mean(lt_p_q[:ind + pad, :])
        r_q_mean = utility.quat_mean(rt_p_q[:ind + pad, :])
    else:
        l_q_mean = utility.quat_mean(lt_p_q[ind - pad:ind + pad, :])
        r_q_mean = utility.quat_mean(rt_p_q[ind - pad:ind + pad, :])

    lt_p_R = utility.quat2matrix(l_q_mean)
    rt_p_R = utility.quat2matrix(r_q_mean)

    # compute the left and right hip coordinate systems
    l_e1 = pelvis_axis
    l_e3 = lt_p_R @ l_thigh_axis
    l_e2 = cross(l_e3, l_e1)
    l_e2 /= norm(l_e2)

    r_e1 = pelvis_axis
    r_e3 = rt_p_R @ r_thigh_axis
    r_e2 = cross(r_e3, r_e1)
    r_e2 /= norm(r_e2)

    # form the pelvis anatomical axes
    p_z = pelvis_axis
    p_x = (r_e2 + l_e2) / 2
    p_y = cross(p_z, p_x)
    p_y /= norm(p_y)
    p_x = cross(p_y, p_z)
    p_x /= norm(p_x)

    # form the left thigh anatomical axes
    lt_x = lt_p_R.T @ l_e2
    lt_y = l_thigh_axis
    lt_z = cross(lt_x, lt_y)
    lt_z /= norm(lt_z)

    # form the right thigh anatomical axes
    rt_x = rt_p_R.T @ r_e2
    rt_y = r_thigh_axis
    rt_z = cross(rt_x, rt_y)
    rt_z /= norm(rt_z)

    # create the anatomical frames
    pelvis_AF = (p_x, p_y, p_z)
    l_thigh_AF = (lt_x, lt_y, lt_z)
    r_thigh_AF = (rt_x, rt_y, rt_z)

    return pelvis_AF, l_thigh_AF, r_thigh_AF



"""
Methods for calculating OMC joint angles

GNU GPL v3.0
Lukas Adamowicz
V0.1 - April 10, 2019
"""
from numpy import cross, sum as nsum, arctan2 as atan2, arccos, abs as nabs, pi, stack
from numpy.linalg import norm


def hip(pelvis_frames, thigh_frames, side):
    """
    Compute hip angles.

    Parameters
    ----------
    pelvis_frames : numpy.ndarray
        Nx3x3 array of anatomical frames of the pelvis in a world frame. N time samples of 3x3 matrices, of which the
        first column is the pelvis X-axis, second column is the pelvis Y-axis, and third column is the pelvis Z-axis
    thigh_frames: numpy.ndarray
        Nx3x3 array of anatomical frames of the thigh for N time points in the world frame. Each 3x3 matrix is comprised
        of columns of thigh x-axis, y-axis, and z-axis, in that order.
    side : {'left', 'right'}
        Side the angles are being computed for. Used for sign determination of angles.

    Returns
    -------
    hip_angles : numpy.ndarray
        Nx3 array of hip angles, with the first column being flexion-extension, second column being ad/abduction,
        and third column being internal-external rotation.

    References
    ----------
    Wu et al. "ISB recommendations on definitions of joint coordinate systems of various joints for the reporting of
    human joint motion - part I: ankle, hip, and spine." J. of Biomech. Vol. 35. 2002.
    Dabirrahmani et al. "Modification of the Grood and Suntay Joint Coordinate System equations for knee joint flexion."
    Med. Eng. and Phys. Vol. 39. 2017.
    Grood et al. "A joint coordinate system for the clinical description of three-dimensional motions: application to
    the knee." J. of Biomech. Engr. Vol. 105. 1983.
    """
    # extract the proximal (pelvis) segment axes
    X = pelvis_frames[:, :, 0]
    Z = pelvis_frames[:, :, 2]

    # extract the distal (thigh) segment axes
    x = thigh_frames[:, :, 0]
    y = thigh_frames[:, :, 1]

    # create the hip joint axes
    e1 = Z.copy()
    e3 = y.copy()

    e2 = cross(e3, e1)
    e2 /= norm(e2, axis=1, keepdims=True)

    # compute the angles by finding the angle between specific axes
    sgn = nsum(cross(X, e2) * Z, axis=1)
    sgn /= nabs(sgn)
    fe = atan2(sgn * norm(cross(X, e2), axis=1), nsum(X * e2, axis=1))

    if side.lower() == 'right':
        sgn = nsum(cross(x, e2) * -y, axis=1)
    elif side.lower() == 'left':
        sgn = nsum(cross(x, e2) * y, axis=1)
    else:
        raise ValueError('Side must be "left" or "right".')

    sgn /= nabs(sgn)
    ier = atan2(sgn * norm(cross(x, e2), axis=1), nsum(x * e2, axis=1))

    if side.lower() == 'right':
        aa = -pi / 2 + arccos(nsum(e1 * e3, axis=1))
    elif side.lower() == 'left':
        aa = pi / 2 - arccos(nsum(e1 * e3, axis=1))

    return stack((fe, aa, ier), axis=1) * 180 / pi


"""
Testing imports of required libraries
"""


def test_numpy():
    import numpy

    return


def test_scipy():
    import scipy

    return


def test_pykinematics():
    import pykinematics
    from pykinematics import imu
    from pykinematics import omc
    from pykinematics.core import MimuAngles, OmcAngles

    return


def test_pykinematics_imu():
    from pykinematics.imu import angles, calibration, joints, optimize, orientation, utility

    return


def test_pykinematics_omc():
    from pykinematics.omc import angles, calibration, segmentFrames, MarkerNames, default_marker_names

    return



"""
Methods for calculating sensor absolute and relative orientations

GNU GPL v3.0
Lukas Adamowicz

V0.1 - March 8, 2019
"""
from numpy import array, zeros, cross, sqrt, abs as nabs, arccos, sin, mean, identity, sum, outer
from numpy.linalg import norm, inv as np_inv

from pykinematics.imu import utility
from pykinematics.imu.optimize import UnscentedKalmanFilter


__all__ = ['MadgwickAHRS', 'OrientationComplementaryFilter', 'SSRO']


class MadgwickAHRS:
    def __init__(self, sample_period=1/256, q_init=array([1, 0, 0, 0]), beta=0.041):
        """
        Algorithm for estimating the orientation of an inertial sensor with or without a magnetometer.

        Parameters
        ----------
        sample_period : float, optional
            Sampling period in seconds.  Default is 1/256 (ie sampling frequency is 256Hz).
        q_init : numpy.ndarray, optional
            Initial quaternion estimate.  Default is [1, 0, 0, 0].
        beta : float, optional
            Beta value for the algorithm.  Default is 1.0

        References
        ----------
        S. Madgwick et al. "Estimation of IMU and MARG orientation using a gradient descent algorithm." *IEEE Intl.
        Conf. on Rehab. Robotics*. 2011.
        """
        self.sample_period = sample_period
        self.q = q_init
        self.beta = beta

    def update(self, gyr, acc, mag):
        """
        Perform update step.

        Parameters
        ----------
        gyr : numpy.ndarray
            Angular velocity at time t.  Units of rad/s.
        acc : numpy.ndarray
            Acceleration at time t.  Units of g.
        mag : numpy.ndarray
            Magnetometer reading at time t.

        Returns
        -------
        q : numpy.ndarray
            Quaternion estimate of orientation at time t.
        """
        # short name for the quaternion
        q = self.q

        # normalize accelerometer measurement
        a = acc / norm(acc)

        # normalize magnetometer measurement
        h = mag / norm(mag)

        # reference direction of earth's magnetic field
        h_ref = utility.quat_mult(q, utility.quat_mult(array([0, h[0], h[1], h[2]]), utility.quat_conj(q)))
        b = array([0, norm(h_ref[1:3]), 0, h_ref[3]])

        # Gradient Descent algorithm corrective step
        F = array([2 * (q[1] * q[3] - q[0] * q[2]) - a[0],
                   2 * (q[0] * q[1] + q[2] * q[3]) - a[1],
                   2 * (0.5 - q[1]**2 - q[2]**2) - a[2],
                   2 * b[1] * (0.5 - q[2]**2 - q[3]**2) + 2 * b[3] * (q[1] * q[3] - q[0] * q[2]) - h[0],
                   2 * b[1] * (q[1] * q[2] - q[0] * q[3]) + 2 * b[3] * (q[0] * q[1] + q[2] * q[3]) - h[1],
                   2 * b[1] * (q[0] * q[2] + q[1] * q[3]) + 2 * b[3] * (0.5 - q[1]**2 - q[2]**2) - h[2]])
        J = array([[-2 * q[2], 2 * q[3], -2 * q[0], 2 * q[1]],
                   [2 * q[1], 2 * q[0], 2 * q[3], 2 * q[2]],
                   [0, -4 * q[1], -4 * q[2], 0],
                   [-2 * b[3] * q[2], 2 * b[3] * q[3], -4 * b[1] * q[2] - 2 * b[3] * q[0], -4 * b[1] * q[3] + 2 * b[3] * q[1]],
                   [-2 * (b[1] * q[3] - b[3] * q[1]), 2 * (b[1] * q[2] + b[3] * q[0]), 2 * (b[1] * q[1] + b[3] * q[3]), -2 * (b[1] * q[0] - b[3] * q[2])],
                   [2 * b[1] * q[2], 2 * b[1] * q[3] - 4 * b[3] * q[1], 2 * b[1] * q[0] - 4 * b[3] * q[2], 2 * b[1] * q[1]]])

        step = J.T @ F
        step /= norm(step)  # normalize step magnitude

        # compute rate of change of quaternion
        qDot = 0.5 * utility.quat_mult(q, array([0, gyr[0], gyr[1], gyr[2]])) - self.beta * step

        # integrate to yeild quaternion
        self.q = q + qDot * self.sample_period
        self.q /= norm(self.q)

        return self.q

    def updateIMU(self, gyr, acc):
        """
        Perform update step using only gyroscope and accelerometer measurements

        Parameters
        ----------
        gyr : numpy.ndarray
            Angular velocity at time t.  Units of rad/s.
        acc : numpy.ndarray
            Acceleration at time t.  Units of g.

        Returns
        -------
        q : numpy.ndarray
            Quaternion estimate of orientation at time t.
        """
        a = acc / norm(acc)  # normalize accelerometer magnitude
        q = self.q  # short name
        # gradient descent algorithm corrective step
        F = array([2 * (q[1] * q[3] - q[0] * q[2]) - a[0],
                   2 * (q[0] * q[1] + q[2] * q[3]) - a[1],
                   2 * (0.5 - q[1]**2 - q[2]**2) - a[2]])
        J = array([[-2 * q[2], 2 * q[3], -2 * q[0], 2 * q[1]],
                   [2 * q[1], 2 * q[0], 2 * q[3], 2 * q[2]],
                   [0, -4 * q[1], -4 * q[2], 0]])
        step = J.T @ F
        step /= norm(step)  # normalise step magnitude

        # compute rate of change quaternion
        q_dot = 0.5 * utility.quat_mult(q, array([0, gyr[0], gyr[1], gyr[2]])) - self.beta * step

        # integrate to yeild quaternion
        q = q + q_dot * self.sample_period
        self.q = q / norm(q)  # normalise quaternion

        return self.q


class OrientationComplementaryFilter:
    def __init__(self, g=9.81, alpha=0.1, beta=0.1, compute_bias=True, bias_alpha=0.1, acc_thresh=0.1, gyr_thresh=0.2,
                 delta_gyr_thresh=0.01, adaptive_gain=True, gain_error_thresh=(0.1, 0.2), interp_thresh=0.9,
                 acc_corr=True):
        """
        Complementary filter for orientation estimation.

        Parameters
        ----------
        g : float, optional
            Local value of gravitational acceleration. Default is 9.8 m/s^2.
        alpha : float, optional
            Acceleration complementary filter cut-off frequency.  Default is 0.1.
        beta : float, optional
            Magnetic field complementary filter cut-off frequency. Default is 0.1.
        compute_bias : bool, optional
            Compute the bias of the angular velocity. Default is True.
        bias_alpha : float, optional
            Angular velocity bias gain.  Default is 0.1
        acc_thresh : float, optional
            Threshold for determining still periods from acceleration for bias updating. Default is 0.1.
        gyr_thresh : float, optional
            Threshold for determining still periods from angular velocity for bias updating. Default is 0.2
        delta_gyr_thresh : float, optional
            Threshold for determining still periods from change since previous update of angular velocity for
            bias updating.  Default is 0.01.
        adaptive_gain : bool, optional
            Use adaptive gain or not.  Default is True.
        gain_error_thresh : array_like, optional
            Cutoff values for gain factor calculation. Default is (0.1, 0.2).
        interp_thresh : float, optional
            Below which value to use SLERP vs LERP. Default is 0.9
        acc_corr : bool, optional
            Correct acceleration by removing rotational acceleration. Default is True

        References
        ----------
        Valenti et al. Keeping a Good Attitude: A Quaternion-Based Orientation Filter for IMU and MARGs. Sensors. 2015
        """
        self.gravity = g

        self.alpha = alpha
        self.beta = beta

        self.comp_bias = compute_bias
        self.bias_alpha = bias_alpha

        self.acc_thresh = acc_thresh
        self.w_thresh = gyr_thresh
        self.delta_w_thresh = delta_gyr_thresh

        self.ada_gain = adaptive_gain
        self.ada_thresh1 = gain_error_thresh[0]
        self.ada_thresh2 = gain_error_thresh[1]
        # setup slope and intercept for adaptive gain factor calculation
        self.ada_m = 1 / (self.ada_thresh1 - self.ada_thresh2)
        self.ada_b = 1 - self.ada_m * self.ada_thresh1

        self.interp_thresh = interp_thresh

        self.acc_corr = acc_corr
        # initialize angular velocity bias
        self.w_bias = zeros(3)

    def run(self, acc, gyr, mag, dt, r):
        """
        Run the orientation estimation over the data

        Parameters
        ----------
        acc : numpy.ndarray
            Nx3 array of accelerations.
        gyr : numpy.ndarray
            Nx3 array of angular velocities.
        mag : numpy.ndarray
            Nx3 array of magnetic field readings.
        dt : float
            Sampling time.

        Attributes
        ----------
        q_ : numpy.ndarray
            Nx4 array of orientation estimates from the global frame to the local frame.
        """
        # store input values
        if self.acc_corr:
            corr = cross(gyr, cross(gyr, r))
            self.a = acc - corr
            # store the acceleration norm since its useful
            self.a_mag = norm(self.a, axis=1)

            self.a /= norm(self.a, axis=1, keepdims=True)
        else:
            # store the acceleration norm since its useful
            self.a_mag = norm(acc, axis=1)
            self.a = acc / self.a_mag.reshape((-1, 1))
        self.w = gyr.copy()
        self.h = mag / norm(mag, axis=1, keepdims=True)
        self.dt = dt

        # get the first estimate of q
        self.q = self._get_measurement(0)

        # get the number of elements
        n = self.a_mag.size

        # storage for quaternion estimate
        self.q_ = zeros((n, 4))

        # run estimation procedure
        for i in range(n):
            # update the bias if doing so
            if self.comp_bias:
                self._update_biases(i)

            # get the prediction for the state
            q_pred = self._get_prediction(i)

            # get the acceleration based correction
            dq_acc = self._get_acc_correction(q_pred, i)

            # get the adaptive gain factor.  Factor is 1 if not using an adaptive gain
            factor = self._get_adaptive_gain_factor(i)
            alpha = self.alpha * factor

            # interpolate the acceleration based correction
            dq_acc_int = self._get_scaled_quaternion(dq_acc, alpha)

            q_pred_acorr = utility.quat_mult(q_pred, dq_acc_int)

            # get the magnetometer based correction
            dq_mag = self._get_mag_correction(q_pred_acorr, i)

            # interpolate the magnetometer based correction
            dq_mag_int = self._get_scaled_quaternion(dq_mag, self.beta)

            # correct the prediction resulting in final estimate
            self.q = utility.quat_mult(q_pred_acorr, dq_mag_int)

            # normalize estimate
            self.q /= norm(self.q)

            # save the orientation estimate
            self.q_[i] = self.q

    def _get_measurement(self, ind):
        """
        Get the initial measurement guess for the algorithm.
        """
        if self.a[ind, 2] >= 0:
            b = sqrt(self.a[ind, 2] + 1)
            q_acc = array([b / sqrt(2), -self.a[ind, 1] / (sqrt(2) * b), self.a[ind, 0] / (sqrt(2) * b), 0])
        else:
            b = sqrt(1 - self.a[ind, 2])
            q_acc = array([-self.a[ind, 1] / (sqrt(2) * b), b / sqrt(2), 0, self.a[ind, 0] / (sqrt(2) * b)])

        l = utility.quat2matrix(q_acc).T @ self.h[ind]
        Gamma = l[0]**2 + l[1]**2

        if l[0] >= 0:
            b = sqrt(Gamma + l[0] * sqrt(Gamma))
            q_mag = array([b / sqrt(2 * Gamma), 0, 0, l[1] / (sqrt(2) * b)])
        else:
            b = sqrt(Gamma - l[0] * sqrt(Gamma))
            q_mag = array([l[1] / (sqrt(2) * b), 0, 0, b / sqrt(2 * Gamma)])

        q_meas = utility.quat_mult(q_acc, q_mag)
        return q_meas

    def _get_state(self, ind):
        """
        Check whether or not a time point is in steady state.

        Parameters
        ----------
        ind : int
            Index of data to check

        Returns
        -------
        steady_state : bool
            Whether or not the sensor is in a steady state.
        """
        if abs(self.a_mag[ind] - self.gravity) > self.acc_thresh:
            return False
        elif (nabs(self.w[ind] - self.w[ind-1]) > self.delta_w_thresh).any():
            return False
        elif (nabs(self.w[ind] - self.w_bias) > self.w_thresh).any():
            return False
        else:
            return True

    def _update_biases(self, ind):
        """
        Update the bias parameters if in steady state.

        Parameters
        ----------
        ind : int
            Index of data to update from
        """
        steady_state = self._get_state(ind)

        if steady_state:
            self.w_bias += self.bias_alpha * (self.w[ind] - self.w_bias)

    def _get_prediction(self, ind):
        """
        Compute the predicted orientation estimate for the time point.

        Parameters
        ----------
        ind : int
            Index of data to use.

        """
        """
        # construct the derivative calculation matrix
        Omega = zeros((4, 4))
        Omega[0, 1:] = self.w[ind] - self.w_bias
        Omega[1:, 0] = -self.w[ind] - self.w_bias
        Omega[1:, 1:] = OrientationComplementaryFilter.skew_symmetric(self.w[ind] - self.w_bias)

        # compute the predicted orientation estimate
        q_pred = self.q + Omega @ self.q * self.dt
        """
        wu = self.w[ind] - self.w_bias

        q_pred = self.q.copy()
        q_pred[0] += 0.5 * self.dt * (wu[0] * self.q[1] + wu[1] * self.q[2] + wu[2] * self.q[3])
        q_pred[1] += 0.5 * self.dt * (-wu[0] * self.q[0] - wu[1] * self.q[3] + wu[2] * self.q[2])
        q_pred[2] += 0.5 * self.dt * (wu[0] * self.q[3] - wu[1] * self.q[0] - wu[2] * self.q[1])
        q_pred[3] += 0.5 * self.dt * (-wu[0] * self.q[2] + wu[1] * self.q[1] - wu[2] * self.q[0])

        return q_pred

    def _get_acc_correction(self, q_pred, ind):
        """
        Compute the acceleration based quaternion correction for the predicted value.

        Parameters
        ----------
        ind : int
            Index to use for computation.

        Returns
        -------
        dq : numpy.ndarray
            Quaternion correction.
        """
        # compute the predicted gravity vector
        gp = utility.quat2matrix(utility.quat_inv(q_pred)) @ self.a[ind]
        gp /= norm(gp)
        # compute the correction quaternion
        # b = sqrt(gp[2] + 1)
        # dq = array([b / sqrt(2), -gp[1] / (sqrt(2) * b), gp[0] / (sqrt(2) * b), 0])
        dq = zeros(4)
        dq[0] = sqrt((gp[2] + 1) * 0.5)
        dq[1] = -gp[1] / (2 * dq[0])
        dq[2] = gp[0] / (2 * dq[0])
        return dq / norm(dq)

    def _get_mag_correction(self, q_pred, ind):
        """
        Compute the magnetometer based correction for the predicted orientation.

        Parameters
        ----------
        q_pred : numpy.ndarray
            Predicted orientation estimate
        ind : int
            Index to compute correction at.

        Returns
        -------
        dq : numpy.ndarray
            Quaternion correction
        """
        # rotate the magnetic field vector into the current orientation estimate
        l = utility.quat2matrix(utility.quat_inv(q_pred)) @ self.h[ind]
        l /= norm(l)
        Gamma = l[0]**2 + l[1]**2

        # b = sqrt(Gamma + l[0] * sqrt(Gamma))
        # dq = array([b / sqrt(2 * Gamma), 0, 0, l[1] / (sqrt(2) * b)])

        beta = sqrt(Gamma + l[0] * sqrt(Gamma))
        dq = zeros(4)
        dq[0] = beta / sqrt(2 * Gamma)
        dq[3] = l[1] / (sqrt(2) * beta)

        return dq / norm(dq)

    def _get_scaled_quaternion(self, q, gain, p=array([1, 0, 0, 0])):
        """
        Scale the quaternion via linear interoplation or spherical linear interpolation.

        Parameters
        ----------
        q : numpy.ndarray
            Quaternion to scale.
        p : numpy.ndarray
            Quaternion to scale towards.

        Returns
        -------
        qs : numpy.ndarray
            Scaled quaternion.
        """
        angle = arccos(sum(q * p))
        if q[0] < self.interp_thresh:
            qs = sin((1 - gain) * angle) / sin(angle) * p + sin(gain * angle) / sin(angle) * q
        else:
            qs = (1 - gain) * p + gain * q

        return qs / norm(qs)

    def _get_adaptive_gain_factor(self, ind):
        """
        Get the adaptive gain factor.

        Parameters
        ----------
        ind : int
            Index to use for gain computation.

        Returns
        -------
        factor : float
            Gain factor.
        """
        factor = 1
        if not self.ada_gain:
            return factor
        else:
            err = abs(self.a_mag[ind] - self.gravity) / self.gravity
            if self.ada_thresh1 < err <= self.ada_thresh2:
                factor = self.ada_m * err + self.ada_b
            elif err > self.ada_thresh2:
                factor = 0
            return factor

    @staticmethod
    def _skew_symmetric(x):
        """
        Get the skew symmetric representation of a vector.

        Parameters
        ----------
        x : numpy.ndarray
            1D vector in 3D space (ie 3 component vector)

        Returns
        -------
        X : numpy.ndarray
            3x3 skew symmetric matrix representation of x
        """
        return array([[0, -x[2], x[1]], [x[2], 0, -x[0]], [-x[1], x[0], 0]])


class SSRO:
    def __init__(self, c=0.04, N=16, error_factor=1e-15, sigma_g=1e-2, sigma_a=1e-1, grav=9.81, init_window=8):
        """
        Sensor-to-Sensor Relative Orientation estimation algorithm.

        Parameters
        ----------
        c : float, optional
            Value between 0 and 1 for the cutoff of the estimation of the gravity vector. Default is 0.04.
        N : int, optional
            Number of samples in the moving average measurement error estimation. Default is 16.
        error_factor : float, optional
            Error factor for the computation of the measurement error for the estimation of the rotation quaternion.
            Default is 1e-15, recommended is to be slightly less than the values in Q, which can be accomplished with
            values around dt**2 * sigma_g**2 * 0.1
        sigma_g : float, optional
            Gyroscope noise value. Default is 0.01
        sigma_a : float, optional
            Accelerometer noise value. Default is 0.1
        grav : float, optional
            Gravitational acceleration. Default is 9.81 m/s^2
        init_window : int, optional
            Number of samples to use for initialization of state and covariance. Default is 8.
        """
        if not 0 <= c <= 1:
            raise ValueError('c must be between 0 and 1')
        self.c = c
        self.N = N
        self.err_factor = error_factor
        self.sigma_g = sigma_g
        self.sigma_a = sigma_a
        self.grav = grav
        self.init_window = init_window

    def run(self, s1_a, s2_a, s1_w, s2_w, s1_m, s2_m, dt):
        """
        Run the SSRO algorithm on the data from two body-segment adjacent sensors. Rotation quaternion
        is from the second sensor to the first sensor.

        Parameters
        ----------
        s1_w : numpy.ndarray
            Sensor 1 angular velocity in rad/s.
        s2_w : numpy.ndarray
            Sensor 2 angular velocity in rad/s.
        s1_a : numpy.ndarray
            Sensor 1 acceleration in m/s^2.
        s2_a : numpy.ndarray
            Sensor 2 acceleration in m/s^2.
        s1_m : numpy.ndarray
            Sensor 1 magnetometer readings.
        s2_m : numpy.ndarray
            Sensor 2 magnetometer readings.
        dt : float, optional
            Sample rate in seconds.

        Returns
        -------
        q : numpy.ndarray
            Rotation quaternion containing the rotation from sensor 2's frame to that of sensor 1's
        """
        # initialize values
        self.x = zeros(10)  # state vector
        g1_init = mean(s1_a[:self.init_window, :], axis=0)
        g2_init = mean(s2_a[:self.init_window, :], axis=0)
        self.x[:3] = g1_init / norm(g1_init)
        self.x[3:6] = g2_init / norm(g2_init)

        m1_init = mean(s1_m[:self.init_window, :], axis=0)
        m2_init = mean(s2_m[:self.init_window, :], axis=0)
        mg1_init = m1_init - sum(m1_init * self.x[0:3]) * self.x[0:3]
        mg2_init = m2_init - sum(m2_init * self.x[3:6]) * self.x[3:6]

        q_g = utility.vec2quat(self.x[3:6], self.x[:3])
        q_m = utility.vec2quat(utility.quat2matrix(q_g) @ mg2_init, mg1_init)
        self.x[6:] = utility.quat_mult(q_m, q_g)

        self.P = identity(10) * 0.01
        self.a1, self.a2 = zeros(3), zeros(3)
        self.a1_, self.a2_ = zeros(s1_w.shape), zeros(s1_w.shape)
        self.eps1, self.eps2 = zeros((3, 3)), zeros((3, 3))

        # storage for the states
        self.x_ = zeros((s1_w.shape[0], 10))

        # run the update step over all the data
        for i in range(s1_w.shape[0]):
            self.update(s1_a[i, :], s2_a[i, :], s1_w[i, :], s2_w[i, :], s1_m[i, :], s2_m[i, :], dt, i)
            self.x_[i] = self.x
            self.a1_[i] = self.a1
            self.a2_[i] = self.a2

        return self.x_

    def update(self, y_a1, y_a2, y_w1, y_w2, y_m1, y_m2, dt, j):
        """
        Update the state vector and covariance for the next time point.
        """
        # short hand names for parts of the state vector
        g1 = self.x[:3]
        g2 = self.x[3:6]
        q_21 = self.x[6:]
        # compute the difference in angular velocities in the second sensor's frame
        R_21 = utility.quat2matrix(q_21)  # rotation from sensor 2 to sensor 1
        y_w1_2 = R_21.T @ y_w1  # sensor 1 angular velocity in sensor 2 frame
        diff_y_w = y_w2 - y_w1_2

        # create the state transition matrix
        A = zeros((10, 10))
        A[0:3, 0:3] = identity(3) - dt * SSRO._skew(y_w1)
        A[3:6, 3:6] = identity(3) - dt * SSRO._skew(y_w2)
        A[6:, 6:] = identity(4) - 0.5 * dt * SSRO._skew_quat(diff_y_w)

        # state transition covariance
        Q = zeros((10, 10))
        Q[0:3, 0:3] = -dt**2 * SSRO._skew(g1) @ (self.sigma_g**2 * identity(3)) @ SSRO._skew(g1)
        Q[3:6, 3:6] = -dt**2 * SSRO._skew(g2) @ (self.sigma_g**2 * identity(3)) @ SSRO._skew(g2)
        Q[6:, 6:] = -dt**2 * SSRO._skew_quat(q_21) @ (self.sigma_g**2 * identity(4)) @ SSRO._skew_quat(q_21)

        # predicted state and covariance
        xhat = A @ self.x
        Phat = A @ self.P @ A.T + Q

        # create the measurements
        zt = zeros(10)
        zt[0:3] = y_a1 - self.c * self.a1
        zt[3:6] = y_a2 - self.c * self.a2
        # quaternion measurement
        q_g = utility.vec2quat(g2, g1)
        mg1 = y_m1 - sum(y_m1 * g1) * g1
        mg2 = y_m2 - sum(y_m2 * g2) * g2
        q_m = utility.vec2quat(utility.quat2matrix(q_g) @ mg2, mg1)

        zt[6:] = utility.quat_mult(q_m, q_g)

        # measurement estimation matrix
        H = zeros((10, 10))
        H[:6, :6] = identity(6) * self.grav
        H[6:, 6:] = identity(4)

        # update estimates of the acceleration variance
        self.eps1 += self.c**2 / self.N * outer(self.a1, self.a1)
        self.eps1 -= self.c**2 / self.N * outer(self.a1_[j - self.N], self.a1_[j - self.N])
        self.eps2 += self.c**2 / self.N * outer(self.a2, self.a2)
        self.eps2 -= self.c**2 / self.N * outer(self.a2_[j - self.N], self.a2_[j - self.N])

        # measurement covariance - essentially how much we trust the measurement vector
        M = zeros((10, 10))
        M[0:3, 0:3] = self.sigma_a**2 * identity(3) + self.eps1
        M[3:6, 3:6] = self.sigma_a**2 * identity(3) + self.eps2
        M[6:, 6:] = identity(4) * self.err_factor * (norm(self.a1) + norm(self.a2))

        # kalman gain and state estimation
        K = Phat @ H.T @ np_inv(H @ Phat @ H.T + M)
        xbar = xhat + K @ (zt - H @ xhat)  # not yet normalized
        self.P = (identity(10) - K @ H) @ Phat

        # normalize the state vector
        self.x[0:3] = xbar[0:3] / norm(xbar[0:3])
        self.x[3:6] = xbar[3:6] / norm(xbar[3:6])
        self.x[6:] = xbar[6:] / norm(xbar[6:])

        # acceleration values
        self.a1 = y_a1 - self.x[0:3] * self.grav
        self.a2 = y_a2 - self.x[3:6] * self.grav

    @staticmethod
    def _skew(v):
        return array([[0, -v[2], v[1]],
                      [v[2], 0, -v[0]],
                      [-v[1], v[0], 0]])

    @staticmethod
    def _skew_quat(v):
        if v.size == 3:
            Q = array([[0, v[0], v[1], v[2]],
                       [-v[0], 0, -v[2], v[1]],
                       [-v[1], v[2], 0, -v[0]],
                       [-v[2], -v[1], v[0], 0]])
        elif v.size == 4:
            Q = array([[v[0], v[1], v[2], v[3]],
                       [-v[1], v[0], -v[3], v[2]],
                       [-v[2], v[3], v[0], -v[1]],
                       [-v[3], -v[2], v[1], v[0]]])
        else:
            raise ValueError("Input can only have 3 or 4 elements")
        return Q




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# these lines allow 1 file to control the version, so only 1 file needs to be updated per version change
fid = open('pykinematics/version.py')
vers = fid.readlines()[-1].split()[-1].strip("\"'")
fid.close()

setuptools.setup(
    name="pykinematics",
    version=vers,
    author="Lukas Adamowicz",
    author_email="lukas.adamowicz95@gmail.com",
    description="Calculation of hip joint angles from wearable inertial sensors and optical motion capture.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/M-SenseResearchGroup/pymotion",  # project url, most likely a github link
    # download_url="https://github.com/M-SenseResearchGroup/pymotion",  # link to where the package can be downloaded, most likely PyPI
    project_urls={
        "Documentation": "https://pykinematics.readthedocs.io/en/latest/"
    },
    include_pacakge_data=False,  # set to True if you have data to package, ie models or similar
    # package_data={'pykinematics.imu.data': ['*.h5']},
    # package_data={'package.module': ['data.csv']},  # if data.csv is in a separate module
    packages=setuptools.find_packages(),  # automatically find required packages
    license='MIT',
    python_requires='>=3.6',  # Version of python required
    install_requires=[],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python :: 3.7"
    ],
)



"""
Core functions for computing joint angles from start to finish

Lukas Adamowicz
May 2019
GNU GPL v3.0
"""
from numpy import mean, diff
from scipy.signal import butter, filtfilt
from warnings import warn

from pykinematics import imu
from pykinematics import omc


class MimuAngles:
    def __init__(self, static_window=1.0, gravity_value=9.81, filter_values=None, angular_velocity_derivative_order=2,
                 joint_center_kwargs=None, orientation_kwargs=None, correct_knee=True, knee_axis_kwargs=None,
                 verbose=True):
        """
        Compute angles from MIMU sensors, from initial raw data through joint angles.

        Parameters
        ----------
        static_window : float, optional
            Window size in seconds for which to use in calibration during static standing. Default is 1s.
        gravity_value : float, optional
            Local gravitational acceleration. Default is 9.81m/s^2
        filter_values : {None, dict}, optional
            Filter values for the inertial and magnetic field data. Default is None, which uses the default settings.
            Providing a dictionary with any of the values modified (ie don't need to specify all 4) will changed the
            specific setting. Each entry is a length 2 tuple, containing first filter order, then cutoff frequency (Hz).
            Default settings and keys:
                - 'Acceleration': (2, 15)
                - 'Angular velocity': (2, 15)
                - 'Angular acceleration': (2, 15)
                - 'Magnetic field': (2, 15)
        angular_velocity_derivative_order : {2, 4}, optional
            Order for the calculation of the angular velocity derivative. Default is 2 for 2nd order.
        joint_center_kwargs : {None, dict}, optional
            Optional joint center computation key-word arguments, or None, for using the defaults. See
            :meth:`pykinematics.imu.joints.Center` for the possible arguments. Default is None.
        orientation_kwargs : {None, dict}, optional
            Optional sensor relative orientation key-word arguments, or NOne, for using the defaults. See
            :meth:`pykinematics.imu.orientation.SSRO` for the possible arguments. Default is None.
        correct_knee : bool, optional
            Correct the knee joint center location by shifting it along the rotation axis closer to the sensors. Default
            is True.
        knee_axis_kwargs : {None, dict}, optional
            Optional knee-axis computation key-word arguments. See :meth:`pykinematics.imu.joints.KneeAxis` for the
            possible arguments. Default is None
        verbose : bool, optional
            Print messages regarding the status of the estimation process. Default is True
        """
        self.static_window = static_window
        self.grav_val = gravity_value

        # set the default filter values
        self.set_default_filter_values()
        if filter_values is not None:  # see if any changes to be included
            if isinstance(filter_values, dict):  # check that the format is followed
                for key in filter_values.keys():  # change the default values
                    self.filt_vals[key] = filter_values[key]

        self.wd_ord = angular_velocity_derivative_order
        assert self.wd_ord == 2 or self.wd_ord == 4, 'The angular velocity derivative order must be either 2 or 4.'

        if joint_center_kwargs is not None:
            self.center_kwargs = joint_center_kwargs
        else:
            self.center_kwargs = {}
        if orientation_kwargs is not None:
            self.orient_kwargs = orientation_kwargs
        else:
            self.orient_kwargs = {}
        if knee_axis_kwargs is not None:
            self.knee_axis_kwargs = knee_axis_kwargs
        else:
            self.knee_axis_kwargs = {}

        self.correct_knee = correct_knee

        # calibrate attributes
        self.acc_scales = None
        self.pelvis_axis, self.l_thigh_axis, self.r_thigh_axis = None, None, None
        self.pelvis_AF, self.l_thigh_AF, self.r_thigh_AF = None, None, None

        self.verbose = verbose

    def estimate(self, task_data, return_orientation=False):
        """
        Estimate joint angles from data during a trial of interest.

        Parameters
        ----------
        task_data : dict
            Dictionary of data from a task to compute the hip joint angles for. Required keys/sensors are Lumbar,
            Left and Right thighs. See the *Notes* for *ImuAngles.calibrate*.
        return_orientation : bool, optional
            Return the estimates of the relative orientation. Default is False.

        Returns
        -------
        left_hip_angles : numpy.ndarray
            (N, 3) array of hip angles for the left hip, in the order Flexion - Extension, Ad / Abduction, and
            Internal - External rotation.
        right_hip_angles : numpy.ndarray
            (N, 3) array of hip angles for the right hip, in the order Flexion - Extension, Ad / Abduction, and
            Internal - External rotation.
        R_Lthigh_lumbar : numpy.ndarray, optional
            Rotations from the left thigh sensor to the lumbar sensor for all time points in the provided data.
            Only returned if `return_orientation` is True.
        R_Rthigh_lumbar : numpy.ndarray, optional
            Rotations from the right thigh sensor to the lumbar sensor for all time points in the provided data.
            Only returned if `return_orientation` is True.
        q_Lthigh_lumbar : numpy.ndarray, optional
            Rotation quaternions from the left thigh sensor to the lumbar sensor for all time points in the provided data.
            Only returned if `return_orientation` is True.
        q_Rthigh_lumbar : numpy.ndarray, optional
            Rotation quaternions from the right thigh sensor to the lumbar sensor for all time points in the provided data.
            Only returned if `return_orientation` is True.
        """
        # check to ensure that the data provided has the required sensors
        MimuAngles._check_required_sensors(task_data, 'trial')

        if self.verbose:
            print('-------------------------------------------------\nPreprocessing trial data...')

        # scale the acceleration data
        for sensor in task_data.keys():
            task_data[sensor]['Acceleration'] *= self.acc_scales[sensor]

        # filter the data
        self._apply_filter_dict(task_data, comp_angular_accel=False)  # don't need angular acceleration

        # compute the relative orientation
        if self.verbose:
            print('Computing sensor relative orientation...')
        srof = imu.orientation.SSRO(grav=self.grav_val, **self.orient_kwargs)
        q_lt_lb, R_lt_lb = MimuAngles._compute_orientation(srof, task_data['Lumbar'], task_data['Left thigh'])
        q_rt_lb, R_rt_lb = MimuAngles._compute_orientation(srof, task_data['Lumbar'], task_data['Right thigh'])

        # compute joint angles
        if self.verbose:
            print('Computing left and right hip joint angles...')
        l_hip_ang = imu.angles.hip_from_frames(self.pelvis_AF, self.l_thigh_AF, R_lt_lb, side='left')
        r_hip_ang = imu.angles.hip_from_frames(self.pelvis_AF, self.r_thigh_AF, R_rt_lb, side='right')

        if return_orientation:
            return l_hip_ang, r_hip_ang, R_lt_lb, R_rt_lb, q_lt_lb, q_rt_lb
        else:
            return l_hip_ang, r_hip_ang

    def calibrate(self, static_data, joint_center_data):
        """
        Calibration by computing the sensor-to-segment alignment.

        Parameters
        ----------
        static_data : dict
            Dictionary of sensor data from a static standing trial. Must contain at least Lumbar, Left thigh, and Right
            thigh keys. Left shank and Right shank are suggested in order to provide an acceleration scale to ensure
            that the measured gravity vectors are equivalent between sensors. See *Notes* for the structure of the
            dictionary.
        joint_center_data : dict
            Dictionary of sensor data from a task to be used for joint center computation. As such the task should have
            sufficient rotation about all possible axes (ie all 3 for the hip, 1 for the knee). Must contain Lumbar,
            thigh, and shank keys. See *Notes*.

        Attributes
        ----------
        self.acc_scales : dict
            Dictionary of acceleration scales for each of the sensors in static_data, required to scale acceleration
            magnitude to that of local gravitational acceleration during static standing
        self.pelvis_axis : numpy.ndarray
            Pelvis fixed axis in the Lumbar sensor's reference frame.
        self.l_thigh_axis : numpy.ndarray
            Left thigh fixed axis in the left thigh's reference frame.
        self.r_thigh_axis : numpy.ndarray
            Right thigh fixed axis in the right thigh's reference frame.
        self.pelvis_AF : tuple
            Tuple of the x, y, and z anatomical axes for the pelvis, in the lumbar sensor frame.
        self.l_thigh_AF : tuple
            Tuple of the x, y, and z anatomical axes for the left thigh, in the left thigh's sensor frame.
        self.r_thigh_AF : tuple
            Tuple of the x, y, and z anatomical axes for the right thigh, in the right thigh's sensor frame.

        Notes
        -----
        Data dictionary key-level schematic:

        - Lumbar
            - Time [s]
            - Acceleration [m/s^2]
            - Angular velocity [rad/s]
            - Magnetic field [not specified]
        - Left thigh
            - Time
            - Acceleration
            - Angular velocity
            - Magnetic field
        - Right thigh
            - Time
            - Acceleration
            - Angular velocity
            - Magnetic field
        - (Left shank)
            - Time
            - Acceleration
            - Angular velocity
            - Magnetic field
        - (Right shank)
            - Time
            - Acceleration
            - Angular velocity
            - Magnetic field
        """
        # check to ensure that the data provided has the required sensors
        MimuAngles._check_required_sensors(static_data, 'static')
        MimuAngles._check_required_sensors(joint_center_data, 'joint center')

        if self.verbose:
            print('\n-----------------------------------------------------\n'
                  'Scaling acceleration and pre-processing the raw data')

        # get the acceleration scales
        self.acc_scales = dict()
        for sensor in static_data.keys():
            self.acc_scales[sensor] = imu.calibration.get_acc_scale(static_data[sensor]['Acceleration'],
                                                                    gravity=self.grav_val)

        # scale the available data
        for sensor in self.acc_scales.keys():
            static_data[sensor]['Acceleration'] *= self.acc_scales[sensor]

            if sensor in list(joint_center_data.keys()):
                joint_center_data[sensor]['Acceleration'] *= self.acc_scales[sensor]
        # issue a warning if any sensors are in the joint center data but not in the static data scales
        for sens in [sens for sens in joint_center_data.keys() if sens not in list(self.acc_scales.keys())]:
            warn(f'Sensor ({sens}) in joint center data has not been scaled due to no scale factor available from '
                 f'static data provided. Performance may suffer as a result.')

        # filter the static data
        self._apply_filter_dict(static_data, comp_angular_accel=False)
        # filter the joint center data
        self._apply_filter_dict(joint_center_data, comp_angular_accel=True)  # need angular accel for this one

        if self.verbose:
            print('Computing joint centers...')
        # compute joint centers from the joint center data
        joint_center = imu.joints.Center(g=self.grav_val, **self.center_kwargs)

        # if the center method is "SAC", we need to compute the relative orientation first
        srof = imu.orientation.SSRO(grav=self.grav_val, **self.orient_kwargs)

        _, jcR_lt_lb = MimuAngles._compute_orientation(srof, joint_center_data['Lumbar'],
                                                       joint_center_data['Left thigh'])
        _, jcR_rt_lb = MimuAngles._compute_orientation(srof, joint_center_data['Lumbar'],
                                                       joint_center_data['Right thigh'])
        _, jcR_ls_lt = MimuAngles._compute_orientation(srof, joint_center_data['Left thigh'],
                                                       joint_center_data['Left shank'])
        _, jcR_rs_rt = MimuAngles._compute_orientation(srof, joint_center_data['Right thigh'],
                                                       joint_center_data['Right shank'])

        # compute the joint centers
        hip_l_lb, hip_l_t, hip_l_res = MimuAngles._compute_center(joint_center, joint_center_data['Lumbar'],
                                                                  joint_center_data['Left thigh'], jcR_lt_lb)
        hip_r_lb, hip_r_t, hip_r_res = MimuAngles._compute_center(joint_center, joint_center_data['Lumbar'],
                                                                  joint_center_data['Right thigh'], jcR_rt_lb)
        knee_l_t, knee_l_s, knee_l_res = MimuAngles._compute_center(joint_center, joint_center_data['Left thigh'],
                                                                    joint_center_data['Left shank'], jcR_ls_lt,
                                                                    self.correct_knee, self.knee_axis_kwargs)
        knee_r_t, knee_r_s, knee_r_res = MimuAngles._compute_center(joint_center, joint_center_data['Right thigh'],
                                                                    joint_center_data['Right shank'], jcR_rs_rt,
                                                                    self.correct_knee, self.knee_axis_kwargs)

        if self.verbose:
            print('------------------------------------------------------------')
            print(
                f'Left hip:  Residual: {hip_l_res:0.3f}\nLumbar: ({hip_l_lb[0] * 100:0.2f}, {hip_l_lb[1] * 100:0.2f}, '
                f'{hip_l_lb[2] * 100:0.2f})cm    Left thigh: ({hip_l_t[0] * 100:0.2f}, {hip_l_t[1] * 100:0.2f}, '
                f'{hip_l_t[2] * 100:0.2f})cm')
            print(
                f'Right hip:  Residual: {hip_r_res:0.3f}\nLumbar: ({hip_r_lb[0] * 100:0.2f}, {hip_r_lb[1] * 100:0.2f}, '
                f'{hip_r_lb[2] * 100:0.2f})cm    Right thigh: ({hip_r_t[0] * 100:0.2f}, {hip_r_t[1] * 100:0.2f}, '
                f'{hip_r_t[2] * 100:0.2f})cm')
            print(
                f'Left knee:  Residual: {knee_l_res:0.3f}\nLeft thigh: ({knee_l_t[0] * 100:0.2f}, '
                f'{knee_l_t[1] * 100:0.2f}, {knee_l_t[2] * 100:0.2f})cm    Left shank: ({knee_l_s[0] * 100:0.2f}, '
                f'{knee_l_s[1] * 100:0.2f}, {knee_l_s[2] * 100:0.2f})cm')
            print(
                f'Right knee:  Residual: {knee_r_res:0.3f}\nRight thigh: ({knee_r_t[0] * 100:0.2f}, '
                f'{knee_r_t[1] * 100:0.2f}, {knee_r_t[2] * 100:0.2f})cm    Right shank: ({knee_r_s[0] * 100:0.2f}, '
                f'{knee_r_s[1] * 100:0.2f}, {knee_r_s[2] * 100:0.2f})cm')
            print('------------------------------------------------------------')
            print('Computing fixed axes and creating anatomical reference frames')

        # compute the fixed axes for the thighs and pelvis
        self.pelvis_axis = imu.joints.fixed_axis(hip_l_lb, hip_r_lb, center_to_sensor=True)
        self.l_thigh_axis = imu.joints.fixed_axis(knee_l_t, hip_l_t, center_to_sensor=True)
        self.r_thigh_axis = imu.joints.fixed_axis(knee_r_t, hip_r_t, center_to_sensor=True)

        # compute the relative orientation between sensors during the static data
        q_lt_lb, _ = MimuAngles._compute_orientation(srof, static_data['Lumbar'], static_data['Left thigh'])
        q_rt_lb, _ = MimuAngles._compute_orientation(srof, static_data['Lumbar'], static_data['Right thigh'])

        # process the static calibration
        AF = imu.calibration.static(q_lt_lb, q_rt_lb, self.pelvis_axis, self.l_thigh_axis, self.r_thigh_axis,
                                    static_data['Lumbar']['Angular velocity'],
                                    static_data['Left thigh']['Angular velocity'],
                                    static_data['Right thigh']['Angular velocity'], static_data['Lumbar']['dt'],
                                    self.static_window)
        self.pelvis_AF, self.l_thigh_AF, self.r_thigh_AF = AF

        if self.verbose:
            print('Calibration complete\n')

    @staticmethod
    def _compute_center(jc, prox, dist, R_dist_prox, correct_knee=False, knee_axis_kwargs=None):
        """
        Compute the joint center

        Parameters
        ----------
        jc : pykinematics.imu.joints.Center
            Initialized joint center computation object
        prox : dict
            Dictionary, containing 'Acceleration', 'Angular velocity', and 'Angular acceleration' readings from a
            sensor.
        dist : dict
            Dictionary, containing 'Acceleration', 'Angular velocity', and 'Angular acceleration' readings from a
            sensor.
        R_dist_prox : numpy.ndarray
            (N, 3, 3) array of rotation matrices that align the distal sensor's frame with the proximal sensor's frame.
        correct_knee : bool, optional
            Whether or not to correct the knee joint. Should only be used for knee joint center computation. Default is
            False.
        knee_axis_kwargs : {None, dict}, optional
            Additional keyword arguments to be passed to the knee axis estimation, which is used for knee joint
            center correction. Default is None.

        Returns
        -------
        prox_joint_center : numpy.ndarray
            Vector from the joint center to the proximal sensor.
        dist_joint_center : numpy.ndarray
            Vector from the joint center to the distal sensor.
        residual : float
            Residual value from the optimization process.
        """
        # run the computation
        prox_jc, dist_jc, res = jc.compute(prox['Acceleration'], dist['Acceleration'],
                                           prox['Angular velocity'], dist['Angular velocity'],
                                           prox['Angular acceleration'], dist['Angular acceleration'], R_dist_prox)
        if correct_knee:
            imu.joints.correct_knee(prox['Angular velocity'], dist['Angular velocity'], prox_jc, dist_jc,
                                    R_dist_prox[0], knee_axis_kwargs)
        return prox_jc, dist_jc, res

    @staticmethod
    def _compute_orientation(sro, sensor1, sensor2):
        """
        Run the orientation estimation filter. Rotation is provided from sensor 2 -> sensor 1

        Parameters
        ----------
        sro : pykinematics.imu.orientation.SROFilter
            Object for estimating the sensor relative orientation
        sensor1 : dict
            Dictionary, containing 'Acceleration', 'Angular velocity', and 'Magnetic field' readings from a sensor.
        sensor2 : dict
            Dictionary, containing 'Acceleration', 'Angular velocity', and 'Magnetic field' readings from a sensor,
            which will be used to find the rotation from sensor2's frame to sensor1's frame

        Returns
        -------
        q_21 : numpy.ndarray
            (N, 4) array quaternions representing the rotation to align sensor2's reference frame with that of sensor1's
        R_21 : numpy.ndarray
            (N, 3, 3) array of rotation matrices corresponding to the quaternions of 'q'
        """
        x = sro.run(sensor1['Acceleration'], sensor2['Acceleration'], sensor1['Angular velocity'],
                    sensor2['Angular velocity'], sensor1['Magnetic field'], sensor2['Magnetic field'], sensor1['dt'])
        # x is the length 10 vector of [gravity_1, gravity_2, q_21]
        R = imu.utility.quat2matrix(x[:, 6:])  # convert to a rotation matrix

        return x[:, 6:], R

    def _apply_filter_dict(self, data, comp_angular_accel=False):
        """
        Apply a filter to a whole dictionary of sensor data, and calculate angular acceleration if necessary.

        Parameters
        ----------
        data : dict
            Dictionary of data to apply the filter to
        comp_angular_accel : bool, optional
            Compute and filter angular acceleration. Default is False
        """
        for sensor in data.keys():
            # compute the sampling time difference, 1/f_sample
            data[sensor]['dt'] = mean(diff(data[sensor]['Time']))
            # apply the specified filter to the acceleration
            data[sensor]['Acceleration'] = MimuAngles._apply_filter(data[sensor]['Acceleration'], data[sensor]['dt'],
                                                                    self.filt_vals['Acceleration'][0],
                                                                    self.filt_vals['Acceleration'][1])
            # apply the specified filter to the angular velocity
            data[sensor]['Angular velocity'] = MimuAngles._apply_filter(data[sensor]['Angular velocity'],
                                                                        data[sensor]['dt'],
                                                                        self.filt_vals['Angular velocity'][0],
                                                                        self.filt_vals['Angular velocity'][1])
            # apply the specified filter to the magnetic field reading
            data[sensor]['Magnetic field'] = MimuAngles._apply_filter(data[sensor]['Magnetic field'],
                                                                      data[sensor]['dt'],
                                                                      self.filt_vals['Magnetic field'][0],
                                                                      self.filt_vals['Magnetic field'][1])

            if comp_angular_accel:
                data[sensor]['Angular acceleration'] = imu.utility.calc_derivative(data[sensor]['Angular velocity'],
                                                                                   data[sensor]['dt'],
                                                                                   order=self.wd_ord)
                data[sensor]['Angular acceleration'] = MimuAngles._apply_filter(data[sensor]['Angular acceleration'],
                                                                                data[sensor]['dt'],
                                                                                self.filt_vals['Angular acceleration'][0],
                                                                                self.filt_vals['Angular acceleration'][1])

    @staticmethod
    def _apply_filter(x, dt, filt_order, filt_cutoff):
        """
        Apply a filter to the data

        Parameters
        ----------
        x : numpy.ndarray
            (N, M) array of data to be filtered. Will filter along the 0th axis (N)
        dt : float
            Sampling time difference between samples.
        filt_order : int
            Order of the filter to be applied
        filt_cutoff : float
            Cutoff frequency of the filter to be applied
        """
        fc = butter(filt_order, filt_cutoff * 2 * dt)
        if x.ndim == 2:
            x = filtfilt(fc[0], fc[1], x, axis=0)
        elif x.ndim == 1:
            x = filtfilt(fc[0], fc[1], x)

        return x

    def set_default_filter_values(self):
        """
        Set the filter values to the default:

        Angular velocity : (2, 15)
        Acceleration : (2, 15)
        Angular acceleration : (2, 15)
        Magnetic field : (2, 15)
        """
        self.filt_vals = {'Angular velocity': (2, 15), 'Acceleration': (2, 15), 'Angular acceleration': (2, 15),
                          'Magnetic field': (2, 15)}

    @staticmethod
    def _check_required_sensors(data, data_use, bilateral=True):
        """
        Check for the required sensors

        Parameters
        ----------
        data : dict
            Dictionary of input data that will be used for estimating joint angles
        data_use : {'static', 'joint center', 'trial'}
            Use for the data. Either 'static', 'joint center', or 'trial'

        Raises
        ------
        ValueError
            If the required sensors are missing from the dictionary
        """
        if data_use == 'static':
            # required sensors : lumbar, left and right thigh
            req = ['Lumbar', 'Left thigh', 'Right thigh']
            if not all([i in [j for j in data.keys()] for i in req]):
                raise ValueError(f'Static data does not have the required sensors. Ensure it has "Lumbar", '
                                 f'"Left thigh", and "Right thigh" data.')
        elif data_use == 'joint center':
            # required sensors : lumbar, left and right thigh, left and right shank
            req = ['Lumbar', 'Left thigh', 'Right thigh', 'Left shank', 'Right shank']
            if not all([i in [j for j in data.keys()] for i in req]):
                raise ValueError(f'Joint center computation data does not have the required sensors. Ensure it has '
                                 f'"Lumbar", "Left thigh", "Right thigh", "Left shank", and "Right shank" data.')
        elif data_use == 'trial':
            # required sensors : lumbar, left and right thigh
            req = ['Lumbar', 'Left thigh', 'Right thigh']
            if not all([i in [j for j in data.keys()] for i in req]):
                raise ValueError(f'Trial data does not have the required sensors. Ensure it has "Lumbar", '
                                 f'"Left thigh", and "Right thigh" data.')


class OmcAngles:
    def __init__(self, marker_names='default', window=1.0):
        """
        Estimation of hip joint angles from optical motion capture marker trajectories.

        Parameters
        ----------
        marker_names : {default, pykinematics.omc.MarkerNames}, optional
            List of marker names and what they are named in the data provided. Default value is `default`, which
            will use the default values, which can be seen through :meth:`pykinematics.omc.default_marker_names`.
        window : float, optional
            Window size in seconds of the duration to use to find the period of best stillness. Default is 1s.
        """
        self.marker_names = marker_names
        self.window_t = window

        # initialize things used later
        self.pelvis = None
        self.left_thigh = None
        self.right_thigh = None

    def estimate(self, task_data, pelvis_cluster=False):
        """
        Estimate hip joint angles from optical motion capture marker trajectories, following calibration.

        Parameters
        ----------
        task_data : dict
            Dictionary of marker position data for a static standing trial, separated based on segment attached to.
            Requires at least Pelvis, Left thigh, and Right thigh. See *Notes* for the layout.
        pelvis_cluster : bool, optional
            Use the pelvis cluster to anatomical frame to create the anatomical frame for the trial, or create the
            pelvis anatomical frame from the marker trajectories (requires that the markers used to create the frame
            be present for the trial).

        Returns
        -------
        left_hip_angles : numpy.ndarray
            Array of hip angles for the left hip, in the order Flexion - Extension, Ad / Abduction, and
            Internal - External rotation.
        right_hip_angles : numpy.ndarray
            Array of hip angles for the right hip, in the order Flexion - Extension, Ad / Abduction, and
            Internal - External rotation.
        """
        # create the segment frames
        pelvis_af = omc.segmentFrames.pelvis(task_data['Pelvis'], use_cluster=pelvis_cluster, R_s_c=self.pelvis[2],
                                             marker_names=self.marker_names)
        lthigh_af = omc.segmentFrames.thigh(task_data['Left thigh'], 'left', use_cluster=True, R_s_c=self.left_thigh[2],
                                            marker_names=self.marker_names)
        rthigh_af = omc.segmentFrames.thigh(task_data['Right thigh'], 'right', use_cluster=True,
                                            R_s_c=self.right_thigh[2],
                                            marker_names=self.marker_names)

        # compute the joint angles
        l_hip_angles = omc.angles.hip(pelvis_af, lthigh_af, 'left')
        r_hip_angles = omc.angles.hip(pelvis_af, rthigh_af, 'right')

        return l_hip_angles, r_hip_angles

    def calibrate(self, static_data, joint_center_data, fs):
        """
        Calibrate optical motion capture segments based on standard definitions of fixed and floating axes.

        Parameters
        ----------
        static_data : dict
            Dictionary of marker position data for a static standing trial, separated based on segment attached to.
            Requires at least Pelvis, Left thigh, and Right thigh. See *Notes* for the layout.
        joint_center_data : dict
            Dictionary of marker position data for a static standing trial, separated based on segment attached to.
            Requires at least Pelvis, Left thigh, and Right thigh. See *Notes* for the layout.
        fs : float
            Sampling frequency in Hz.

        Attributes
        ----------
        self.pelvis
            3-tuple of pelvis anatomical frame, rotation from world to pelvis cluster frame, and constant rotation from
            segment frame to cluster frame.
        self.left_thigh : tuple
            3-tuple of left thigh anatomical frame, rotation from world to the left thigh cluster frame, and constant
            rotation from segment frame to cluster frame.
        self.right_thigh : tuple
            3-tuple of right thigh anatomical frame, rotation from world to the right thigh cluster frame, and constant
            rotation from segment frame to cluster frame.

        Notes
        -----
        Data dictionary key-level schematic:

        - Pelvis
            - left_asis
            - right_asis
            - ...
        - Left thigh
            - left_lat_femoral_epicondyle
            - ...
        - Right thigh
            - right_lat_femoral_epicondyle
            - ...
        """
        # TODO update omc.calibration.process_static to accept dictionaries. Or change the whole input
        # change to dumb format specified originally
        static_data_tuple = (static_data['Pelvis'], static_data['Left thigh'], static_data['Right thigh'])
        joint_center_data_tuple = (joint_center_data['Pelvis'], joint_center_data['Left thigh'],
                                   joint_center_data['Right thigh'])

        window = int(self.window_t * fs)
        # process the calibration data
        self.pelvis, self.left_thigh, self.right_thigh = omc.calibration.static(static_data_tuple,
                                                                                joint_center_data_tuple,
                                                                                window, self.marker_names)



"""
Methods that are common and needed accross both IMU and OMC methods

GNU GPL v3.0
Lukas Adamowicz
V0.1 - April 10, 2019
"""
from numpy import array, zeros, argmin, ceil, mean, std, sqrt, nanmean, nanstd


def mov_avg(seq, window):
    """
    Compute the centered moving average and standard deviation

    Parameters
    ----------
    seq : numpy.ndarray
        Data to take the moving average and standard deviation on.
    window : int
        Window size for the moving average/standard deviation.

    Returns
    -------
    m_mn : numpy.ndarray
        Moving average
    m_st : numpy.ndarray
        Moving standard deviation
    pad : int
        Padding at beginning of the moving average and standard deviation
    """
    m_mn = zeros(seq.shape)
    m_st = zeros(seq.shape)

    if window < 2:
        window = 2

    pad = int(ceil(window / 2))

    # compute first window stats
    if pad < seq.shape[0]:
        m_mn[pad] = mean(seq[:window], axis=0)
        m_st[pad] = std(seq[:window], axis=0, ddof=1)**2  # ddof of 1 indicates sample standard deviation, no population

        # compute moving mean and standard deviation
        for i in range(1, seq.shape[0] - window):
            diff_fl = seq[window + i - 1] - seq[i - 1]  # difference in first and last elements of sliding window
            m_mn[pad + i] = m_mn[pad + i - 1] + diff_fl / window
            m_st[pad + i] = m_st[pad + i - 1] + (seq[window + i - 1] - m_mn[pad + i] + seq[i - 1] -
                                                 m_mn[pad + i - 1]) * diff_fl / (window - 1)
    else:
        pad = 1
        m_mn[:] = nanmean(seq, axis=0)
        m_st[:] = nanstd(seq, axis=0, ddof=1)**2

    m_st = sqrt(abs(m_st))  # added absolute value to ensure that any round off error doesn't effect the results

    if window % 2 == 1:
        # m_mn = m_mn[:-1]
        # m_st = m_st[:-1]
        pass
    return m_mn, m_st, pad


def find_most_still(data, window_size, return_index=False):
    """
    Find the most still window of given data by using the minimum of the summed variances for each window.  Returns
    the mean data during the most still window.

    Parameters
    ----------
    data : tuple
        Tuple of data to use for determining the minimum summed variance for each window
    window_size : int
        Number of samples per window.
    return_index : bool
        Return the index of the lowest variance.

    Returns
    -------
    still_data : tuple
        Tuple containing the mean data for each provided data stream for the most still window.
    ind : int, optional
        Index of the lowest variance window, excluding the padding at the start and end of the moving windows.
    """
    var = []
    means = tuple()
    still_data = tuple()
    for d in data:
        m_mn, m_st, pad = mov_avg(d, window_size)

        m_st = m_st ** 2  # square the standard deviation to get the variance

        means += (m_mn[pad:-pad], )
        var.append(m_st[pad:-pad].sum(axis=1))

    ind = argmin(array(var).sum(axis=0))

    for mn in means:
        still_data += (mn[ind], )

    if return_index:
        return still_data, ind
    else:
        return still_data





"""
Utility methods for OMC joint angle estimation

GNU GPL v3.0
Lukas Adamowicz
V0.1 - April 10, 2019
"""
from numpy import cross, stack
from numpy.linalg import norm
from dataclasses import dataclass


@dataclass
class MarkerNames:
    """
    Data class for storing the labels of specific markers needed to calculate hip joint angles. Each class attribute
    is the name of a specific marker location, and its value is the name of that marker in the optical motion capture
    data used to estimate hip joint angles.

    To get a list of the markers, use :meth:`pykinematics.omc.default_marker_names`.
    """
    # pelvis markers
    left_asis: str = 'left_asis'
    right_asis: str = 'right_asis'
    left_psis: str = 'left_psis'
    right_psis: str = 'right_psis'

    # pelvis cluster markers
    pelvis_c1: str = 'sacrum_cluster1'
    pelvis_c2: str = 'sacrum_cluster2'
    pelvis_c3: str = 'sacrum_cluster3'

    # left thigh markers
    left_lep: str = 'left_lat_femoral_epicondyle'
    left_mep: str = 'left_med_femoral_epicondyle'

    # left thigh cluster markers
    left_thigh_c1: str = 'left_thigh_cluster1'
    left_thigh_c2: str = 'left_thigh_cluster2'
    left_thigh_c3: str = 'left_thigh_cluster3'

    # right thigh markers
    right_lep: str = 'right_lat_femoral_epicondyle'
    right_mep: str = 'right_med_femoral_epicondyle'

    # right thigh cluster markers
    right_thigh_c1: str = 'right_thigh_cluster1'
    right_thigh_c2: str = 'right_thigh_cluster2'
    right_thigh_c3: str = 'right_thigh_cluster3'


def default_marker_names():
    """
    Print and return the default marker names.

    Returns
    -------
    marker_names : pykinematics.omc.MarkerNames
        Object for storing the marker names and their associated label in the optical motion capture marker data.
    """
    for key in MarkerNames.__dataclass_fields__.keys():
        print(f'{key}: {MarkerNames.__dataclass_fields__[key].name}')

    return MarkerNames


def create_cluster_frame(marker_data, segment_name, marker_names='default'):
    """
    Create a cluster reference frame for the specified segment.

    Parameters
    ----------
    marker_data : dict
        Dictionary of marker data, where keys are the marker names.
    segment_name : {'pelvis', 'left_thigh', 'right_thigh', 'left_shank', 'right_shank'}
        Name of the segment to calculate the cluster for. Must match a prefix of a cluster name in marker_names
    marker_names : {'default', MarkerNames}, optional
        Either 'default' which will use the default marker names (see MarkerNames class), or a modified MarkerNames
        object where the names are changed to match those in the marker_data keys.

    Returns
    -------
    R_w_c : numpy.ndarray
        Either a 3x3 or Nx3x3 array of rotation matrices of the rotation from the world frame to the cluster frame.
        The first, second, and third columns are also the cluster frame x, y, and z axes respectively.
    """
    # get the marker names to use
    if marker_names == 'default':
        names = MarkerNames()
    else:
        names = marker_names

    # create the cluster names
    name_c1 = segment_name + '_c1'
    name_c2 = segment_name + '_c2'
    name_c3 = segment_name + '_c3'

    # create the cluster frame
    x = marker_data[names.__dict__[name_c3]] - marker_data[names.__dict__[name_c2]]
    x /= norm(x) if x.ndim == 1 else norm(x, axis=1, keepdims=True)

    y_tmp = marker_data[names.__dict__[name_c1]] - marker_data[names.__dict__[name_c2]]

    z = cross(x, y_tmp)
    z /= norm(z) if z.ndim == 1 else norm(z, axis=1, keepdims=True)

    y = cross(z, x)
    y /= norm(y) if y.ndim == 1 else norm(y, axis=1, keepdims=True)

    # create and return the matrix of axes
    R_w_c = stack((x, y, z), axis=1) if x.ndim == 1 else stack((x, y, z), axis=2)
    return R_w_c


def compute_pelvis_origin(left_asis, right_asis):
    """
    Compute the origin of the pelvis.

    Parameters
    ----------
    left_asis : numpy.ndarray
        1 or 2D (Nx3) array of positions of the left asis marker.
    right_asis : numpy.ndarray
        1 or 2D (Nx3) array of positions of the right asis marker.

    Returns
    -------
    origin : numpy.ndarray
        Array of the position of the pelvis origin.

    References
    ----------
    Ren et al.  "Whole body inverse dynamics over a complete gait cycle based only on measured kinematics."
    *J. of Biomech*. 2008
    """
    origin = (left_asis + right_asis) / 2

    return origin


def compute_thigh_origin(lat_fem_ep, med_fem_ep):
    """
    Compute the origin of the thigh.

    Parameters
    ----------
    lat_fem_ep : numpy.ndarray
        1 or 2D (Nx3) array of positions of the lateral femoral epicondyle marker.
    med_fem_ep : numpy.ndarray
        1 or 2D (Nx3) array of positions of the medial femoral epicondyle marker.

    Returns
    -------
    origin : numpy.ndarray
        Array of the position of the thigh origin

    References
    ----------
    Ren et al.  "Whole body inverse dynamics over a complete gait cycle based only on measured kinematics."
    *J. of Biomech*. 2008
    """
    origin = (lat_fem_ep + med_fem_ep) / 2

    return origin


def compute_shank_origin(lat_mall, med_mall):
    """
    Compute the origin of the shank.

    Parameters
    ----------
    lat_mall : numpy.ndarray
        1 or 2D (Nx3) array of positions of the lateral malleolus marker.
    med_mall : numpy.ndarray
        1 or 2D (Nx3) array of positions of the medial malleolus marker.

    Returns
    -------
    origin : numpy.ndarray
        Array of the position of the pelvis origin.

    References
    ----------
    Ren et al.  "Whole body inverse dynamics over a complete gait cycle based only on measured kinematics."
    *J. of Biomech*. 2008
    """
    origin = (lat_mall + med_mall) / 2

    return origin



"""
Testing of functions in pykinematics/common.py
"""
from pykinematics.common import *

import numpy as np


def test_mov_avg_window_1(simple_x):
    _, _, pad = mov_avg(simple_x, 1)

    assert pad == 1


def test_mov_avg_window_large(simple_x):
    mn, sd, pad = mov_avg(simple_x, len(simple_x) * 2 + 4)

    assert pad == 1
    assert np.allclose(mn, np.mean(simple_x))
    assert np.allclose(sd, np.std(simple_x, ddof=1))


def test_mov_avg(simple_x, ma_window, simple_x_mov_stats):
    x_m, x_sd, pad = mov_avg(simple_x, ma_window)

    assert np.allclose(x_m, simple_x_mov_stats[0])
    assert np.allclose(x_sd, simple_x_mov_stats[1])
    assert np.isclose(pad, 3)


def test_find_most_still():
    np.random.seed(5)
    x1 = np.random.rand(50, 2)
    np.random.seed(34)
    x2 = np.random.rand(50, 2)
    np.random.seed(13513)
    x3 = np.random.rand(50, 2)

    still_data, idx = find_most_still((x1, x2, x3), 10, return_index=True)
    result = find_most_still((x1, x2, x3), 10, return_index=False)

    assert np.allclose(np.array(still_data), np.array([[0.49851287, 0.39499911], [0.60036786, 0.45481188],
                                                       [0.18751978, 0.54213798]]))
    assert np.isclose(idx, 7)
    assert len(result) == 3  # should be 3 because thats the length of the input, which is the shape of the return
